#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy
import tensorflow

# Import usienarl

from usienarl import Agent, Interface, SpaceType
from usienarl.models import DeepDeterministicPolicyGradient


class RandomAgent(Agent):
    """
    Random agents taking random actions in all modes.
    It does not require warm-up.

    Note: no save/restore is ever performed by this agent.
    """

    def __init__(self,
                 name: str):
        # Generate base agent
        super(RandomAgent, self).__init__(name)

    def setup(self,
              logger: logging.Logger,
              scope: str,
              parallel: int,
              observation_space_type: SpaceType, observation_space_shape: (),
              agent_action_space_type: SpaceType, agent_action_space_shape: (),
              summary_path: str = None, save_path: str = None, saves_to_keep: int = 0) -> bool:
        # Make sure parameters are correct
        assert(parallel > 0)
        logger.info("Setup of agent " + self._name + " with scope " + scope + "...")
        # Reset agent attributes
        self._scope = scope
        self._parallel = parallel
        self._observation_space_type: SpaceType = observation_space_type
        self._observation_space_shape = observation_space_shape
        self._agent_action_space_type: SpaceType = agent_action_space_type
        self._agent_action_space_shape = agent_action_space_shape
        # Use a blank generate method
        if not self._generate(logger,
                              observation_space_type, observation_space_shape,
                              agent_action_space_type, agent_action_space_shape):
            return False
        # Validate setup
        return True

    def restore(self,
                logger: logging.Logger,
                session,
                path: str) -> bool:
        return True

    def save(self,
             logger: logging.Logger,
             session):
        pass

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape: (),
                  agent_action_space_type: SpaceType, agent_action_space_shape: ()) -> bool:
        return True

    def initialize(self,
                   logger: logging.Logger,
                   session):
        pass

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current: numpy.ndarray,
                   warmup_step: int, warmup_episode: int):
        pass

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current: numpy.ndarray,
                  train_step: int, train_episode: int):
        # Act randomly
        return interface.sample_agent_action(logger, session)

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current: numpy.ndarray,
                      inference_step: int, inference_episode: int):
        # Act randomly
        return interface.sample_agent_action(logger, session)

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current: numpy.ndarray,
                             agent_action: numpy.ndarray,
                             reward: numpy.ndarray,
                             episode_done: numpy.ndarray,
                             agent_observation_next: numpy.ndarray,
                             warmup_step: int, warmup_episode: int):
        pass

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current: numpy.ndarray,
                            agent_action: numpy.ndarray,
                            reward: numpy.ndarray,
                            episode_done: numpy.ndarray,
                            agent_observation_next: numpy.ndarray,
                            train_step: int, train_episode: int):
        pass

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current: numpy.ndarray,
                                agent_action: numpy.ndarray,
                                reward: numpy.ndarray,
                                episode_done: numpy.ndarray,
                                agent_observation_next: numpy.ndarray,
                                inference_step: int, inference_episode: int):
        pass

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: numpy.ndarray,
                                episode_total_reward: numpy.ndarray,
                                warmup_step: int, warmup_episode: int):
        pass

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: numpy.ndarray,
                               episode_total_reward: numpy.ndarray,
                               train_step: int, train_episode: int):
        pass

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: numpy.ndarray,
                                   episode_total_reward: numpy.ndarray,
                                   inference_step: int, inference_episode: int):
        pass

    @property
    def saved_variables(self):
        return None

    @property
    def warmup_steps(self) -> int:
        return 0



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import enum
import logging
import numpy

# Define space types for observation and action spaces definition in the environment


class SpaceType(enum.Enum):
    """
    Space types for environment definition.

    If discrete, each value is defined by an integer.
    If continuous, each value is defined by a vector.
    """
    discrete = 1
    continuous = 2


class Environment:
    """
    Base environment abstract class. It is not ready to be used before setup.

    Actions, states, rewards and episode-done flags are all defined in a vectorized way to allow batch parallelization.
    The parallel amount if set during setup.

    You should always define your own environment, the base class cannot be used as-is.
    """

    def __init__(self,
                 name: str):
        # Make sure parameters are valid
        assert(name != "")
        # Define internal attributes
        self._name: str = name
        # Define empty attributes
        self._parallel: int or None = None

    def setup(self,
              logger: logging.Logger,
              parallel: int = 1) -> bool:
        """
        Setup the environment.

        :param logger: the logger used to print the environment information, warnings and errors
        :param parallel: the amount of parallel episodes run by the environment
        :return: True if the setup is successful, false otherwise
        """
        # Make sure parameters are valid
        assert(parallel > 0)
        # Set parallel attribute
        self._parallel = parallel
        # Generate the environment and check if generation is successful
        return self._generate(logger)

    def _generate(self,
                  logger: logging.Logger) -> bool:
        """
        Generate the environment.

        :param logger: the logger used to print the environment information, warnings and errors
        :return: True if the environment generation is successful, false otherwise
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def initialize(self,
                   logger: logging.Logger,
                   session):
        """
        Initialize the environment.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def close(self,
              logger: logging.Logger,
              session):
        """
        Close the environment (e.g. the window of a rendered OpenAI environment). Usually is not required.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def reset(self,
              logger: logging.Logger,
              session) -> numpy.ndarray:
        """
        Reset the environment to its default state. Requires the environment to be initialized.
        This is used when an episode is done to start a new episode.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        :return: the current state of the environment on its default state, wrapped in a numpy array
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def step(self,
             logger: logging.Logger,
             session,
             action: numpy.ndarray) -> ():
        """
        Execute a step on the environment with the given action.
        Requires the environment to be initialized.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        :param action: the action executed on the environment, wrapped in a numpy array
        :return: the next state, the reward obtained and a boolean flag indicating whether or not the episode is done, all wrapped in numpy arrays wrapped in a tuple
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def render(self,
               logger: logging.Logger,
               session):
        """
        Render the environment.
        Note: to render with text, it's advised to use print statement instead of the logger.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def sample_action(self,
                      logger: logging.Logger,
                      session) -> numpy.ndarray:
        """
        Sample a random action from the environment.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        :return: a random action of the environment wrapped in a numpy array
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    def possible_actions(self,
                         logger: logging.Logger,
                         session) -> []:
        """
        Get a list of all actions' indexes possible at the current states of the environment if the action space
        is discrete.
        Get a list containing the lower and the upper bounds at the current states of the environment, each
        wrapped in numpy array with the shape of the  action space.

        :param logger: the logger used to print the environment information, warnings and errors
        :param session: the session of tensorflow currently running
        :return: a list of indices containing the possible actions or a list of upper and lower bounds arrays
        """
        # Abstract method, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def name(self) -> str:
        """
        The name of the environment.
        """
        return self._name

    @property
    def parallel(self) -> int:
        """
        The amount of parallelization of the environment.
        """
        return self._parallel

    @property
    def state_space_type(self) -> SpaceType:
        """
        The type of the state space of the environment.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def state_space_shape(self) -> ():
        """
        The shape of the state space of the environment, wrapped in a tuple.
        Note: it may differ from the agent's observation space shape.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def action_space_type(self) -> SpaceType:
        """
        The type of the action space of the environment.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def action_space_shape(self) -> ():
        """
        The shape of the action space of the environment, wrapped in a tuple.
        Note: it may differ from the agent's action space shape.
        """
        # Abstract property, it should be implemented on a child class basis
        raise NotImplementedError()



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy
import tensorflow

# Import usienarl

from usienarl import Agent, Interface, SpaceType
from usienarl.models import ProximalPolicyOptimization


class PPOAgent(Agent):
    """
    Proximal Policy Optimization agent without specific exploration policies (beside sampling which is automatically
    done by the model). It is supplied with a PPO model.

    The number of episodes to play before updating is a crucial parameter to stabilize learning.
    Increase it if the environment has high variance. The default value of 100 is tailored to training volleys of length
    around 1000.
    Note that the number of episodes actually executed between two updates of the model may actually differ a bit from
    that value with respect to the parallelization used.
    """

    def __init__(self,
                 name: str,
                 model: ProximalPolicyOptimization,
                 update_every_episodes: int = 100):
        # Define internal attributes
        self._model: ProximalPolicyOptimization = model
        self._update_every_episodes: int = update_every_episodes
        # Define empty attributes
        self._last_update_episode: int = 0
        self._current_value_estimate = None
        self._current_log_likelihood = None
        # Generate base agent
        super(PPOAgent, self).__init__(name)

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape: (),
                  agent_action_space_type: SpaceType, agent_action_space_shape: ()) -> bool:
        # Generate the model and check if it's successful, stop if not successful
        return self._model.generate(logger, self._scope + "/" + self._name,
                                    self._parallel,
                                    observation_space_type, observation_space_shape,
                                    agent_action_space_type, agent_action_space_shape)

    def initialize(self,
                   logger: logging.Logger,
                   session):
        # Reset attributes
        self._last_update_episode = 0
        self._current_value_estimate = None
        self._current_log_likelihood = None
        # Initialize the model
        self._model.initialize(logger, session)

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current: numpy.ndarray,
                   warmup_step: int, warmup_episode: int):
        pass

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current: numpy.ndarray,
                  train_step: int, train_episode: int):
        # Act according to the model best prediction (a sample from the probability distribution)
        # Note: it is inherently exploring
        action, self._current_value_estimate, self._current_log_likelihood = self._model.sample_action(session, agent_observation_current, interface.possible_agent_actions(logger, session))
        # Return the exploration action
        return action

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current: numpy.ndarray,
                      inference_step: int, inference_episode: int):
        # Predict the action with the model by sampling from its probability distribution
        action, _, _ = self._model.sample_action(session, agent_observation_current, interface.possible_agent_actions(logger, session))
        # Return the predicted action
        return action

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current: numpy.ndarray,
                             agent_action: numpy.ndarray,
                             reward: numpy.ndarray,
                             episode_done: numpy.ndarray,
                             agent_observation_next: numpy.ndarray,
                             warmup_step: int, warmup_episode: int):
        pass

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current: numpy.ndarray,
                            agent_action: numpy.ndarray,
                            reward: numpy.ndarray,
                            episode_done: numpy.ndarray,
                            agent_observation_next: numpy.ndarray,
                            train_step: int, train_episode: int):
        # Save the current step in the buffer together with the current value estimate and the log-likelihood
        self._model.buffer.store(agent_observation_current.copy(), agent_action.copy(), reward.copy(), episode_done.copy(), self._current_value_estimate.copy(), self._current_log_likelihood.copy())

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current: numpy.ndarray,
                                agent_action: numpy.ndarray,
                                reward: numpy.ndarray,
                                episode_done: numpy.ndarray,
                                agent_observation_next: numpy.ndarray,
                                inference_step: int, inference_episode: int):
        pass

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: numpy.ndarray,
                                episode_total_reward: numpy.ndarray,
                                warmup_step: int, warmup_episode: int):
        pass

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: numpy.ndarray,
                               episode_total_reward: numpy.ndarray,
                               train_step: int, train_episode: int):
        # Update the buffer at the end of trajectory
        self._model.buffer.finish_trajectory(last_step_reward)
        # Execute update only after a certain number of trajectories each time
        if train_episode - self._last_update_episode >= self._update_every_episodes:
            # Execute the update and store policy and value loss
            buffer: [] = self._model.buffer.get()
            logger.info("Updating the model with a buffer of " + str(len(buffer[0])) + " samples")
            policy_stream_loss, value_stream_loss, approximated_kl_divergence, approximated_entropy, clip_fraction = self._model.update(session, buffer)
            # Generate the tensorflow summary
            summary = tensorflow.Summary()
            summary.value.add(tag="policy_stream_loss", simple_value=policy_stream_loss)
            summary.value.add(tag="value_stream_loss", simple_value=value_stream_loss)
            summary.value.add(tag="approximated_kl_divergence", simple_value=approximated_kl_divergence)
            summary.value.add(tag="approximated_entropy", simple_value=approximated_entropy)
            summary.value.add(tag="clip_fraction", simple_value=clip_fraction)
            # Update the summary at the absolute current step if required
            if self._summary_writer is not None:
                self._summary_writer.add_summary(summary, train_step)
            # Update last update episode
            self._last_update_episode = train_episode

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: numpy.ndarray,
                                   episode_total_reward: numpy.ndarray,
                                   inference_step: int, inference_episode: int):
        pass

    @property
    def saved_variables(self):
        return self._model.trainable_variables

    @property
    def warmup_steps(self) -> int:
        return self._model.warmup_steps

    @property
    def update_every_episodes(self) -> int:
        """
        The number of trajectories to play before updating.
        """
        return self._update_every_episodes



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import numpy
import logging

# Import required src

from usienarl import Environment, Volley, Agent, Interface


class StepVolley(Volley):
    """
    Step based volley. Used for warmup volleys.
    """
    def __init__(self,
                 environment: Environment,
                 agent: Agent,
                 interface: Interface,
                 parallel: int,
                 steps_required: int, episode_length: int):
        # Generate base volley
        super(StepVolley, self).__init__(environment, agent, interface, parallel)
        # Make sure additional parameters are correct
        assert (steps_required > 0 and episode_length > 0)
        # Define internal attributes
        self._environment: Environment = environment
        self._agent: Agent = agent
        self._interface: Interface = interface
        self._parallel: int = parallel
        self._steps_required: int = steps_required
        self._episode_length: int = episode_length
        # Define empty attributes
        self._rewards: [] = []
        self._last_episode_done: numpy.ndarray or None = None
        self._last_reward: numpy.ndarray or None = None

    def _initialize(self) -> bool:
        # Reset empty attributes
        self._last_episode_done = None
        self._last_reward = None
        self._rewards = []
        # This initialization always succeed
        return True

    def run(self,
            logger: logging.Logger,
            session,
            render: bool = False):
        # Get the amount of parallel step batches
        parallel_step_batches: int = self._steps_required // self._parallel
        executed_parallel_step_batches: int = 0
        # Execute warm-up
        while True:
            # Initialize last reward and last episode done flags
            self._last_episode_done = numpy.zeros(self._environment.parallel, dtype=bool)
            self._last_reward = numpy.nan * numpy.ones(self._environment.parallel, dtype=float)
            # Execute actions until the all parallel step batches are completed or the maximum episode length is exceeded
            episode_rewards: [] = []
            state_current: numpy.ndarray = self._environment.reset(logger, session)
            for parallel_step_batch in range(self._episode_length):
                # Check if we can stop warming up
                if self._steps > self._steps_required:
                    return
                # Print current progress every once in a while (if length is not too short)
                if parallel_step_batches >= 1000 and executed_parallel_step_batches % (parallel_step_batches // 10) == 0 and executed_parallel_step_batches > 0:
                    logger.info("Warmed-up for " + str(self._steps) + "/" + str(self._steps_required) + " steps...")
                # Get the action decided by the agent
                observation_current: numpy.ndarray = self._interface.environment_state_to_observation(logger, session, state_current)
                agent_action: numpy.ndarray = self._agent.act_warmup(logger, session, self._interface, observation_current,
                                                                     self._start_steps + self._steps, self._start_episodes + self._episodes)
                # Get the next state with relative reward and completion flag
                environment_action: numpy.ndarray = self._interface.agent_action_to_environment_action(logger, session, agent_action)
                state_next, reward, episode_done = self._environment.step(logger, session, environment_action)
                # Send back information to the agent
                observation_next: numpy.ndarray = self._interface.environment_state_to_observation(logger, session, state_next)
                # Complete the step
                self._agent.complete_step_warmup(logger, session, self._interface,
                                                 observation_current, agent_action, reward, episode_done, observation_next,
                                                 self._start_steps + self._steps, self._start_episodes + self._episodes)
                # Render if required
                if render:
                    self._environment.render(logger, session)
                # Save the reward at the last step
                self._last_reward = numpy.where(episode_done * (1 - self._last_episode_done), reward, self._last_reward)
                if parallel_step_batch + 1 == self._episode_length:
                    self._last_reward = numpy.where(numpy.isnan(self._last_reward), reward, self._last_reward)
                # Add the reward to the list of rewards for this episode
                # Note: make sure the reward saved is NaN for all parallel episodes already completed
                episode_rewards.append(numpy.where(self._last_episode_done, numpy.nan, reward))
                # Update the current state with the previously next state
                state_current = state_next.copy()
                # Increase the number of warm-up steps
                # Note: the counter should be increased according to the completed episodes of the current parallel set
                self._steps += numpy.count_nonzero(episode_done == 0)
                # Increase the episode counter
                self._episodes += numpy.count_nonzero((episode_done * (1 - self._last_episode_done)) == 1)
                # Save the current episode done flags
                # Note: episode done flag is a numpy array so it should be copied
                self._last_episode_done = episode_done.copy()
                # Increase the counter of parallel warm-up steps batch already executed globally
                executed_parallel_step_batches += 1
                # Check if the episode is completed
                if all(episode_done):
                    break
            # Consider done also all parallel episodes truncated but not terminated
            self._episodes += numpy.count_nonzero(self._last_episode_done == 0)
            # Save rewards per episode
            self._rewards.append(episode_rewards.copy())
            # Complete the episode and send back information to the agent
            self._agent.complete_episode_warmup(logger, session, self._interface,
                                                self._last_reward, numpy.nansum(numpy.array(episode_rewards), axis=0),
                                                self._start_steps + self._steps, self._start_episodes + self._episodes)

    @property
    def last_episode_done(self) -> numpy.ndarray or None:
        """
        The episode done flag in the last step of the last episode of the volley.
        It is wrapped in a numpy array.
        It is None if volley is not setup.
        """
        return self._last_episode_done

    @property
    def last_reward(self) -> numpy.ndarray or None:
        """
        The reward in the last step of the last episode of the volley.
        It is wrapped in a numpy array.
        It is None if volley is not setup.
        """
        return self._last_reward

    @property
    def rewards(self) -> []:
        """
        The list of rewards per step grouped by episode of all episodes already executed.
        """
        return self._rewards



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import tensorflow
import numpy
import logging
import matplotlib.pyplot as plot
import matplotlib.ticker as mticker

# Import required src

from usienarl import Environment, Agent, Interface
from usienarl.volleys import EpisodeVolley, StepVolley, EpisodeVolleyType
from usienarl.interfaces import PassThroughInterface


class Experiment:
    """
    Base experiment abstract class.
    An experiment is a set of commands which allows to train or test a certain agent on a certain environment using a
    certain interface to translate states to observations and actions between the two.
    If an interface is not supplied, a default simple pass-through interface is used.
    An experiment can be used to train or test an agent. Both training and testing require setup.

    An experiment training phase consists of the following:
        - a warm-up volley, where the agent acts in warm-up mode for a certain amount of steps (usually required to fill
        an internal buffer of the agent model). This is used only by some models and it is run whether the
        agent requires it or not.
        - multiple set of training/validation volleys, where the agent acts in train/inference mode for a certain amount
        of episodes. Training volleys episodes and validation volleys episodes can be set separately, but for each
        training volley a corresponding validation volley is always run. The number of training/validation volleys
        define the length of a training phase.
    During training plots of most important metrics (average of total and scaled reward, standard deviation of total and
    scaled reward, average episode length) are saved with respect to each episode (on a volley basis) and with respect
    to each volley. An experiment training phase terminates when the number of required training/validation volleys is
    reached or when the agent is validated.

    An experiment testing phase consists of the following:
        - a set of test volleys, where the agent acts in inference mode. No plots are saved during test, but additional
        metrics are computed. Test volleys also set an attribute defining whether or not the agent is successful in
        overcoming the test phase.

    To run an experiment you must define your own. To define your own experiment, implement the abstract class in a
    specific child class. A child class only requires to have implemented the methods used to validate and test the
    agent's model and conclude the training and test phase respectively.
    """

    def __init__(self,
                 name: str,
                 environment: Environment,
                 agent: Agent,
                 interface: Interface = None):
        # Define internal attributes
        self._name: str = name
        self._environment: Environment = environment
        self._agent: Agent = agent
        self._interface: Interface = interface if interface is not None else PassThroughInterface(self._environment)
        # Define global empty attributes
        self._plots_path: str or None = None
        self._plots_dpi: int or None = None
        self._tensorflow_config = None
        self._id: str or None = None
        self._parallel: int or None = None
        # Define data empty attributes
        self._warmup_volley: StepVolley or None = None
        self._training_volley: EpisodeVolley or None = None
        self._validation_volley: EpisodeVolley or None = None
        self._training_avg_total_rewards: [] or None = None
        self._training_avg_scaled_rewards: [] or None = None
        self._training_std_total_rewards: [] or None = None
        self._training_std_scaled_rewards: [] or None = None
        self._training_avg_episode_lengths: [] or None = None
        self._validation_avg_total_rewards: [] or None = None
        self._validation_avg_scaled_rewards: [] or None = None
        self._validation_std_total_rewards: [] or None = None
        self._validation_std_scaled_rewards: [] or None = None
        self._validation_avg_episode_lengths: [] or None = None
        self._training_validation_volley_counter: int or None = None
        self._trained_steps: int or None = None
        self._trained_episodes: int or None = None
        self._validated: bool or None = None
        self._test_volley: EpisodeVolley or None = None
        self._test_avg_total_rewards: [] or None = None
        self._test_avg_scaled_rewards: [] or None = None
        self._test_std_total_rewards: [] or None = None
        self._test_std_scaled_rewards: [] or None = None
        self._test_avg_episode_lengths: [] or None = None
        self._test_volley_counter: int or None = None
        self._avg_test_avg_total_reward: float or None = None
        self._max_test_avg_total_reward: float or None = None
        self._avg_test_avg_scaled_reward: float or None = None
        self._max_test_avg_scaled_reward: float or None = None
        self._avg_test_std_total_reward: float or None = None
        self._min_test_std_total_reward: float or None = None
        self._avg_test_std_scaled_reward: float or None = None
        self._min_test_std_scaled_reward: float or None = None
        self._avg_test_avg_episode_length: int or None = None
        self._successful: bool or None = None

    def setup(self,
              logger: logging.Logger,
              plots_path: str, plots_dpi: int = 150,
              parallel: int = 1,
              summary_path: str = None, save_path: str = None, saves_to_keep: int = 0,
              iteration: int = -1) -> bool:
        """
        Setup the experiment, preparing all of its component to execution. This must be called before both train and
        test. Until the experiment is not setup, most of its properties will return None.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param plots_path: the string path of the plots to save (both for the entire experiment and each one of its volleys)
        :param plots_dpi: the dpi of each plot graph (reduce it to reduce plots size and quality)
        :param parallel: the amount of parallel episodes run for the experiment, must be greater than zero
        :param summary_path: the optional string path of the tensorboard summary directory to save during model training
        :param save_path: the optional string path of the metagraph agent directory to save at the end of each train volley
        :param saves_to_keep: the optional number of metagraphs to keep when saved, it does nothing if there is no save path
        :param iteration: number to append to the experiment name in all scopes and print statements (if not less than zero)
        :return: True if the setup of the experiment is successful, False otherwise
        """
        # Make sure parameters are valid
        assert(parallel > 0)
        assert(plots_dpi > 0)
        # Set attributes
        self._id = self._name + "_" + str(iteration) if iteration > -1 else self._name
        self._parallel = parallel
        self._plots_path = plots_path
        self._plots_dpi = plots_dpi
        # Start setup process
        logger.info("Setup of experiment " + self._id + "...")
        # Reset the tensorflow graph to a blank new graph
        tensorflow.reset_default_graph()
        # Execute the setup on the environment and on the agent and return if not both of them setup correctly
        logger.info("Environment setup...")
        if not self._environment.setup(logger, self._parallel):
            logger.info("Environment setup failed. Cannot setup the experiment!")
            return False
        logger.info("Environment setup is successful")
        logger.info("Agent setup...")
        if not self._agent.setup(logger,
                                 self._id, self._parallel,
                                 self._interface.observation_space_type, self._interface.observation_space_shape,
                                 self._interface.agent_action_space_type, self._interface.agent_action_space_shape,
                                 summary_path, save_path + "/" + self._id, saves_to_keep):
            logger.info("Agent setup failed. Cannot setup the experiment!")
            return False
        logger.info("Agent setup is successful")
        # Configure tensorflow GPU
        # Note: allow growth in memory
        self._tensorflow_config = tensorflow.ConfigProto()
        self._tensorflow_config.gpu_options.allow_growth = True
        logger.info("Tensorflow configuration is successful")
        # Configure matplotlib
        # Note: make the matplotlib use agg-backend to make it compatible with a server and increase the chunksize to avoid unwanted errors
        plot.switch_backend('agg')
        plot.rcParams['agg.path.chunksize'] = 1000
        logger.info("Matplotlib configuration is successful")
        # If setup of both environment and agent is successful, reset internal attributes
        # Note: from now on most attributes are accessible
        self._warmup_volley = None
        self._training_volley = None
        self._validation_volley = None
        self._training_avg_total_rewards = []
        self._training_avg_scaled_rewards = []
        self._training_std_total_rewards = []
        self._training_std_scaled_rewards = []
        self._training_avg_episode_lengths = []
        self._validation_avg_total_rewards = []
        self._validation_avg_scaled_rewards = []
        self._validation_std_total_rewards = []
        self._validation_std_scaled_rewards = []
        self._validation_avg_episode_lengths = []
        self._training_validation_volley_counter = 0
        self._trained_steps = 0
        self._trained_episodes = 0
        self._validated = False
        self._test_volley = None
        self._test_avg_total_rewards = []
        self._test_avg_scaled_rewards = []
        self._test_std_total_rewards = []
        self._test_std_scaled_rewards = []
        self._test_avg_episode_lengths = []
        self._test_volley_counter = 0
        self._avg_test_avg_total_reward = None
        self._max_test_avg_total_reward = None
        self._avg_test_avg_scaled_reward = None
        self._max_test_avg_scaled_reward = None
        self._avg_test_std_total_reward = None
        self._min_test_std_total_reward = None
        self._avg_test_std_scaled_reward = None
        self._min_test_std_scaled_reward = None
        self._avg_test_avg_episode_length = None
        self._successful = False
        # Setup is successful
        logger.info("Experiment setup is successful")
        return True

    def _save_plots(self,
                    episode_volley_type: EpisodeVolleyType):
        """
        Save training or validation plots according to the given type. If type is test it does nothing.

        :param episode_volley_type: the type of episode volley to save the appropriate plots. Test is not valid and ignored
        """
        # If test volley avoid saving plots
        if episode_volley_type == EpisodeVolleyType.test:
            return
        # Save plots according to the requested type
        if episode_volley_type == EpisodeVolleyType.training:
            plot.plot(list(range(len(self._training_avg_total_rewards))), self._training_avg_total_rewards, 'r-')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel("Training volley")
            plot.ylabel("Average total reward")
            plot.savefig(self._plots_path + "/training_volleys_total_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(self._training_avg_scaled_rewards))), self._training_avg_scaled_rewards, 'r--')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel("Training volley")
            plot.ylabel("Average scaled reward")
            plot.savefig(self._plots_path + "/training_volleys_scaled_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(self._training_std_total_rewards))), self._training_std_total_rewards, 'c-')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel("Training volley")
            plot.ylabel("Standard deviation of total reward")
            plot.savefig(self._plots_path + "/training_volleys_std_total_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(self._training_std_scaled_rewards))), self._training_std_scaled_rewards, 'c--')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel('Training volley')
            plot.ylabel('Standard deviation of scaled reward')
            plot.savefig(self._plots_path + "/training_volleys_std_scaled_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(self._training_avg_episode_lengths))), self._training_avg_episode_lengths, 'b-.')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel('Training volley')
            plot.ylabel('Average episode length (steps)')
            plot.savefig(self._plots_path + "/training_volleys_average_episode_length.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
        else:
            plot.plot(list(range(len(self._validation_avg_total_rewards))), self._validation_avg_total_rewards, 'm-')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel('Validation volley')
            plot.ylabel('Average total reward')
            plot.savefig(self._plots_path + "/validation_volleys_total_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(self._validation_avg_scaled_rewards))), self._validation_avg_scaled_rewards, 'm--')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel('Validation volley')
            plot.ylabel('Average scaled reward')
            plot.savefig(self._plots_path + "/validation_volleys_scaled_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(self._validation_std_total_rewards))), self._validation_std_total_rewards, 'y-')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel('Validation volley')
            plot.ylabel('Standard deviation of total reward')
            plot.savefig(self._plots_path + "/validation_volleys_std_total_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(self._validation_std_scaled_rewards))), self._validation_std_scaled_rewards, 'y--')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel('Validation volley')
            plot.ylabel('Standard deviation of scaled reward')
            plot.savefig(self._plots_path + "/validation_volleys_std_scaled_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(self._validation_avg_episode_lengths))), self._validation_avg_episode_lengths, 'b-.')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            plot.xlabel('Validation volley')
            plot.ylabel('Average episode length (steps)')
            plot.savefig(self._plots_path + "/validation_volleys_average_episode_length.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()

    def train(self,
              logger: logging.Logger,
              training_volleys_episodes: int, validation_volleys_episodes: int, volleys: int,
              episode_length: int,
              render_training: bool = False, render_validation: bool = False,
              restore_path: str = None):
        """
        Train the experiment's agent for a certain number of episodes per training/validation volley and a certain
        number of volleys. Training and validation volleys are always run in sequence (training first then validation)
        and in the same amount.

        If a restore path is given the model is restored from that (according to the way the agent restores its inner
        model), otherwise a new model is used. Also if restoration fails a new model is used.

        This generates a tensorflow session and initialize both environment and agent associated with the experiment.
        After this is run all training/validation results are accessible as properties.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param training_volleys_episodes: the number of training episodes per training volley, must be greater than zero
        :param validation_volleys_episodes: the number of validation episodes per validation volley, must be greater than zero
        :param volleys: the number of training/validation volleys, must be greater than zero
        :param episode_length: the max length in steps of each episode, must be greater than zero
        :param render_training: boolean flag to render or not the environment during training volleys execution
        :param render_validation: boolean flag to render or not the environment during validation volleys execution
        :param restore_path: the path of the model to restore is stored, if None or empty a new model is built
        """
        # Make sure parameters are valid
        assert(training_volleys_episodes > 0 and validation_volleys_episodes > 0 and volleys > 0 and episode_length > 0)
        # Start experiment training process
        logger.info("Experiment " + self._id + ": training process started")
        # Define the session
        with tensorflow.Session(config=self._tensorflow_config) as session:
            # Initialize the environment and the agent
            logger.info("Initializing the environment...")
            self._environment.initialize(logger, session)
            logger.info("Environment initialized")
            logger.info("Initializing the agent..")
            self._agent.initialize(logger, session)
            logger.info("Agent initialized")
            # Restore the agent if possible
            if restore_path is not None and restore_path:
                logger.info("Restoring agent graph with checkpoint path: " + restore_path)
                if self._agent.restore(logger, session, restore_path):
                    logger.info("Agent restored")
            # Execute warm-up if the agents require so
            if self._agent.warmup_steps > 0:
                # Generate warm-up volley
                self._warmup_volley = StepVolley(self._environment, self._agent, self._interface, self._parallel,
                                                 self._agent.warmup_steps, episode_length)
                # Setup and run warm-up volley
                logger.info("Setup of warm-up volley...")
                if not self._warmup_volley.setup():
                    logger.error("Warm-up volley failed to setup. Exiting...")
                    return
                logger.info("Setup of warm-up volley completed")
                self._warmup_volley.run(logger, session)
                logger.info("Warm-up volley is completed")
            # Generate training and validation volleys
            self._training_volley = EpisodeVolley(self._environment, self._agent, self._interface, self._parallel,
                                                  EpisodeVolleyType.training, self._plots_path, self._plots_dpi,
                                                  training_volleys_episodes, episode_length)
            self._validation_volley = EpisodeVolley(self._environment, self._agent, self._interface, self._parallel,
                                                    EpisodeVolleyType.validation, self._plots_path, self._plots_dpi,
                                                    validation_volleys_episodes, episode_length)
            # Execute training/validation volleys
            start: int = self._training_validation_volley_counter
            for volley_index in range(start, volleys):
                # Setup and run training volley
                logger.info("_____________________________________________________________________________________")
                logger.info("Setup of training volley " + str(self._training_validation_volley_counter) + "...")
                if not self._training_volley.setup(self._training_validation_volley_counter,
                                                   self._trained_steps, self._trained_episodes):
                    logger.error("Training volley failed to setup. Exiting...")
                    return
                logger.info("Setup of training volley " + str(self._training_validation_volley_counter) + " completed")
                self._training_volley.run(logger, session, render_training)
                # Update counter of trained steps and episodes
                self._trained_steps += self._training_volley.steps
                self._trained_episodes += self._training_volley.episodes
                # Store training volley statistics
                self._training_avg_total_rewards.append(self._training_volley.avg_total_reward)
                self._training_avg_scaled_rewards.append(self._training_volley.avg_scaled_reward)
                self._training_std_total_rewards.append(self._training_volley.std_total_reward)
                self._training_std_scaled_rewards.append(self._training_volley.std_scaled_reward)
                self._training_avg_episode_lengths.append(self._training_volley.avg_episode_length)
                # Print training volley results
                logger.info("Training volley " + str(self._training_validation_volley_counter) + " is completed")
                logger.info("Training steps up to now: " + str(self._trained_steps))
                # Save/update plots for all training volleys up to now (except at first volley, just one volley is not enough to plot anything)
                if self._training_validation_volley_counter > 0:
                    if self._training_validation_volley_counter == 0:
                        logger.info("Save all training volleys plots...")
                    else:
                        logger.info("Update all training volleys plots...")
                    self._save_plots(EpisodeVolleyType.training)
                    if self._training_validation_volley_counter == 0:
                        logger.info("Plots saved successfully")
                    else:
                        logger.info("Plots updated successfully")
                logger.info("_____________________________________________________________________________________")
                # Setup and run validation volley
                logger.info("---------------------------------------------------------------------------------------")
                logger.info("Setup of validation volley " + str(self._training_validation_volley_counter) + "...")
                if not self._validation_volley.setup(self._training_validation_volley_counter):
                    logger.error("Validation volley failed to setup. Exiting...")
                    return
                logger.info("Setup of validation volley " + str(self._training_validation_volley_counter) + " completed")
                self._validation_volley.run(logger, session, render_validation)
                # Store validation volley statistics
                self._validation_avg_total_rewards.append(self._validation_volley.avg_total_reward)
                self._validation_avg_scaled_rewards.append(self._validation_volley.avg_scaled_reward)
                self._validation_std_total_rewards.append(self._validation_volley.std_total_reward)
                self._validation_std_scaled_rewards.append(self._validation_volley.std_scaled_reward)
                self._validation_avg_episode_lengths.append(self._validation_volley.avg_episode_length)
                # Print validation volley results
                logger.info("Validation volley " + str(self._training_validation_volley_counter) + " is completed")
                # Save/update plots for all training volleys up to now (except at first volley, just one volley is not enough to plot anything)
                if self._training_validation_volley_counter > 0:
                    if self._training_validation_volley_counter == 0:
                        logger.info("Save all validation volleys plots...")
                    else:
                        logger.info("Update all validation volleys plots...")
                    self._save_plots(EpisodeVolleyType.validation)
                    if self._training_validation_volley_counter == 1:
                        logger.info("Plots saved successfully")
                    else:
                        logger.info("Plots updated successfully")
                logger.info("---------------------------------------------------------------------------------------")
                # Save the agent metagraph
                self._agent.save(logger, session)
                # Update the training/validation volley counter
                self._training_validation_volley_counter += 1
                # Check for validation
                self._validated = self._is_validated(logger)
                if self._validated:
                    logger.info("Validation of the agent is successful")
                    break
            logger.info("End of training")

    def test(self,
             logger: logging.Logger,
             episodes: int, volleys: int,
             episode_length: int,
             restore_path: str,
             render: bool = True):
        """
        Test the experiment's agent whose model is defined with the given checkpoint metagraph for a certain number
        of episodes per volley and a certain number of volleys.

        This generates a tensorflow session and initialize both environment and agent associated with the experiment.
        After this is run all test results are accessible as properties.

        :param logger: the logger used to print the experiment information, warnings and errors
        :param episodes: the number of episodes for each test volley, must be greater than zero
        :param volleys: the number of test volleys, must be greater than zero
        :param episode_length: the max length in steps of each episode, must be greater than zero
        :param restore_path: the path where the metagraph of the model to test is stored
        :param render: boolean flag to render or not the environment
        """
        # Make sure parameters are valid
        assert(episodes > 0 and volleys > 0 and episode_length > 0)
        # Start experiment testing process
        logger.info("Experiment " + self._id + ": testing process started")
        # Define the session
        with tensorflow.Session(config=self._tensorflow_config) as session:
            # Initialize the environment and the agent
            logger.info("Initializing the environment...")
            self._environment.initialize(logger, session)
            logger.info("Environment initialized")
            logger.info("Initializing the agent..")
            self._agent.initialize(logger, session)
            logger.info("Agent initialized")
            # Load the pre-trained model at given checkpoint path
            if restore_path is None or not restore_path:
                logger.error("A checkpoint path is required to test!")
                return
            # Restore the agent model
            logger.info("Restoring agent graph with checkpoint path: " + restore_path)
            if not self._agent.restore(logger, session, restore_path):
                return
            logger.info("Agent restored")
            # Generate test volley
            self._test_volley = EpisodeVolley(self._environment, self._agent, self._interface, self._parallel,
                                              EpisodeVolleyType.test, self._plots_path, self._plots_dpi,
                                              episodes, episode_length)
            # Execute test volleys
            start: int = self._test_volley_counter
            for volley_index in range(start, volleys):
                # Setup and run test volley
                logger.info("Setup of test volley " + str(self._test_volley_counter) + "...")
                if not self._test_volley.setup():
                    logger.error("Test volley failed to setup. Exiting...")
                    return
                self._test_volley.run(logger, session, render)
                # Store test volley statistics
                self._test_avg_total_rewards.append(self._test_volley.avg_total_reward)
                self._test_avg_scaled_rewards.append(self._test_volley.avg_scaled_reward)
                self._test_std_total_rewards.append(self._test_volley.std_total_reward)
                self._test_std_scaled_rewards.append(self._test_volley.std_scaled_reward)
                self._test_avg_episode_lengths.append(self._test_volley.avg_episode_length)
                # Increase test volley counter
                self._test_volley_counter += 1
            logger.info("End of test")
            # Store the average and the best total and scaled rewards over all test volleys
            self._avg_test_avg_total_reward = numpy.round(numpy.average(self._test_avg_total_rewards), 3)
            self._max_test_avg_total_reward = numpy.round(numpy.max(self._test_avg_total_rewards), 3)
            self._avg_test_avg_scaled_reward = numpy.round(numpy.average(self._test_avg_scaled_rewards), 3)
            self._max_test_avg_scaled_reward = numpy.round(numpy.max(self._test_avg_scaled_rewards), 3)
            # Store the average and the minimum standard deviation of total and scaled rewards over test volleys
            self._avg_test_std_total_reward = numpy.round(numpy.average(self._test_std_total_rewards), 3)
            self._min_test_std_total_reward = numpy.round(numpy.min(self._test_std_total_rewards), 3)
            self._avg_test_std_scaled_reward = numpy.round(numpy.average(self._test_std_scaled_rewards), 3)
            self._min_test_std_scaled_reward = numpy.round(numpy.min(self._test_std_scaled_rewards), 3)
            # Store the average episode length over all test volleys
            self._avg_test_avg_episode_length = numpy.rint(numpy.average(self._test_avg_episode_lengths))
            # Print test results
            logger.info("Results over " + str(volleys) + " test volleys of " + str(episodes) + " episodes each are:")
            logger.info("Average test total reward: " + str(self._avg_test_avg_total_reward))
            logger.info("Average test scaled reward: " + str(self._avg_test_avg_scaled_reward))
            logger.info("Max test total reward: " + str(self._max_test_avg_total_reward))
            logger.info("Max test scaled reward: " + str(self._max_test_avg_scaled_reward))
            logger.info("Average test standard deviation of total reward: " + str(self._avg_test_std_total_reward))
            logger.info("Average test standard deviation of scaled reward: " + str(self._avg_test_std_scaled_reward))
            logger.info("Min test standard deviation of total reward: " + str(self._min_test_std_total_reward))
            logger.info("Min test standard deviation of scaled reward: " + str(self._min_test_std_scaled_reward))
            logger.info("Average test episode length: " + str(self._avg_test_avg_episode_length))
            # Check if test is successful
            self._successful = self._is_successful(logger)
            if self._successful:
                logger.info("The experiment terminated successfully")
            else:
                logger.info("The experiment terminated without meeting the requirements")

    def _is_validated(self,
                      logger: logging.Logger) -> bool:
        """
        Check if the experiment training phase is to be considered validated.

        :param logger: the logger used to print the experiment information, warnings and errors
        :return: True if training is validated, False otherwise
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    def _is_successful(self,
                       logger: logging.Logger) -> bool:
        """
        Check if the experiment testing phase is to be considered successful.

        :param logger: the logger used to print the experiment information, warnings and errors
        :return: True if test is successful, False otherwise
        """
        # Abstract method, definition should be implemented on a child class basis
        raise NotImplementedError()

    @property
    def name(self) -> str:
        """
        The name of the experiment.
        """
        return self._name

    @property
    def agent(self) -> Agent:
        """
        The agent associated with the experiment.
        """
        return self._agent

    @property
    def environment(self) -> Environment:
        """
        The environment associated with the experiment.
        """
        return self._environment

    @property
    def interface(self) -> Interface:
        """
        The interface associated with the experiment.
        """
        return self._interface

    @property
    def id(self) -> str or None:
        """
        The id of the experiment, consisting of the name and the iteration number (if greater or equal than zero).
        It is None if experiment is not setup.
        """
        return self._id

    @property
    def plots_path(self) -> str or None:
        """
        The path of plots saved by the experiment and its volleys.
        It is None if experiment is not setup.
        """
        return self._plots_path

    def plots_dpi(self) -> int or None:
        """
        The dpi (quality) of plots saved by the experiment and its volleys.
        It is None if experiment is not setup.
        """
        return self._plots_dpi

    @property
    def parallel(self) -> int or None:
        """
        The amount of parallel episodes run by the environment and the agent of the experiment.
        It is None if experiment is not setup.
        """
        return self._parallel

    @property
    def warmup_volley(self) -> StepVolley or None:
        """
        The warm-up volley of the training phase.
        It is None if warm-up is not required by the agent or not started yet.
        """
        return self._warmup_volley

    @property
    def training_volley(self) -> EpisodeVolley or None:
        """
        The training volley of the training phase.
        It is None if training is not started yet.
        """
        return self._training_volley

    @property
    def validation_volley(self) -> EpisodeVolley or None:
        """
        The validation volley of the training phase.
        It is None if training is not started yet.
        """
        return self._validation_volley

    @property
    def training_avg_total_rewards(self) -> [] or None:
        """
        The list of average total rewards of each volley of all the training volleys already executed.
       It is None if experiment is not setup.
        It is empty if no training volley has finished execution.
        """
        return self._training_avg_total_rewards

    @property
    def training_avg_scaled_rewards(self) -> [] or None:
        """
        The list of average scaled rewards of each volley of all the training volleys already executed.
        It is None if experiment is not setup.
        It is empty if no training volley has finished execution.
        """
        return self._training_avg_scaled_rewards

    @property
    def training_std_total_rewards(self) -> [] or None:
        """
        The list of standard deviations of total rewards of each volley of all the training volleys already executed.
        It is None if experiment is not setup.
        It is empty if no training volley has finished execution.
        """
        return self._training_std_total_rewards

    @property
    def training_std_scaled_rewards(self) -> [] or None:
        """
        The list of standard deviations of scaled rewards of each volley of all the training volleys already executed.
        It is None if experiment is not setup.
        It is empty if no training volley has finished execution.
        """
        return self._training_std_scaled_rewards

    @property
    def training_avg_episode_lengths(self) -> [] or None:
        """
        The list of average episode lengths of each volley of all the training volleys already executed.
        It is None if experiment is not setup.
        It is empty if no training volley has finished execution.
        """
        return self._training_avg_episode_lengths

    @property
    def validation_avg_total_rewards(self) -> [] or None:
        """
        The list of average total rewards of each volley of all the validation volleys already executed.
        It is None if experiment is not setup.
        It is empty if no validation volley has finished execution.
        """
        return self._validation_avg_total_rewards

    @property
    def validation_avg_scaled_rewards(self) -> [] or None:
        """
        The list of average scaled rewards of each volley of all the validation volleys already executed.
        It is None if experiment is not setup.
        It is empty if no validation volley has finished execution.
        """
        return self._validation_avg_scaled_rewards

    @property
    def validation_std_total_rewards(self) -> [] or None:
        """
        The list of standard deviations of total rewards of each volley of all the validation volleys already executed.
        It is None if experiment is not setup.
        It is empty if no validation volley has finished execution.
        """
        return self._validation_std_total_rewards

    @property
    def validation_std_scaled_rewards(self) -> [] or None:
        """
        The list of standard deviations of scaled rewards of each volley of all the validation volleys already executed.
        It is None if experiment is not setup.
        It is empty if no validation volley has finished execution.
        """
        return self._validation_std_scaled_rewards

    @property
    def validation_avg_episode_lengths(self) -> [] or None:
        """
        The list of average episode lengths of each volley of all the validation volleys already executed.
        It is None if experiment is not setup.
        It is empty if no validation volley has finished execution.
        """
        return self._validation_avg_episode_lengths

    @property
    def test_avg_total_rewards(self) -> [] or None:
        """
        The list of average total rewards of each volley of all the test volleys already executed.
        It is None if experiment is not setup.
        It is empty if no test volley has finished execution.
        """
        return self._test_avg_total_rewards

    @property
    def test_avg_scaled_rewards(self) -> [] or None:
        """
        The list of average scaled rewards of each volley of all the test volleys already executed.
        It is None if experiment is not setup.
        It is empty if no test volley has finished execution.
        """
        return self._test_avg_scaled_rewards

    @property
    def test_std_total_rewards(self) -> [] or None:
        """
        The list of standard deviations of total rewards of each volley of all the test volleys already executed.
        It is None if experiment is not setup.
        It is empty if no test volley has finished execution.
        """
        return self._test_std_total_rewards

    @property
    def test_std_scaled_rewards(self) -> [] or None:
        """
        The list of standard deviations of scaled rewards of each volley of all the test volleys already executed.
        It is None if experiment is not setup.
        It is empty if no test volley has finished execution.
        """
        return self._test_std_scaled_rewards

    @property
    def test_avg_episode_lengths(self) -> [] or None:
        """
        The list of average episode lengths of each volley of all the test volleys already executed.
        It is None if experiment is not setup.
        It is empty if no test volley has finished execution.
        """
        return self._test_avg_episode_lengths

    @property
    def trained_steps(self) -> int or None:
        """
        Trained steps, globally, up to now.
        It is None if experiment is not setup.
        """
        return self._trained_steps

    @property
    def trained_episodes(self) -> int or None:
        """
        Trained episodes, globally, up to now.
        It is None if experiment is not setup.
        """
        return self._trained_episodes

    @property
    def training_validation_volley_counter(self) -> int or None:
        """
        Current counter of training/validation volleys.
        It is None if experiment is not setup.
        """
        return self._training_validation_volley_counter

    @property
    def test_volley_counter(self) -> int or None:
        """
        Current counter of test volleys.
        It is None if experiment is not setup.
        """
        return self._test_volley_counter

    @property
    def avg_test_avg_total_reward(self) -> float or None:
        """
        The average total reward among all test volleys.
        It is None if test phase has not finished execution.
        """
        return self._avg_test_avg_total_reward

    @property
    def max_test_avg_total_reward(self) -> float or None:
        """
        The max average total reward among all test volleys.
        It is None if test phase has not finished execution.
        """
        return self._max_test_avg_total_reward

    @property
    def avg_test_avg_scaled_reward(self) -> float or None:
        """
        The average scaled reward among all test volleys.
        It is None if test phase has not finished execution.
        """
        return self._avg_test_avg_scaled_reward

    @property
    def max_test_avg_scaled_reward(self) -> float or None:
        """
        The max average scaled reward among all test volleys.
        It is None if test phase has not finished execution.
        """
        return self._max_test_avg_scaled_reward

    @property
    def avg_test_std_total_reward(self) -> float or None:
        """
        The average standard deviation total reward among all test volleys.
        It is None if test phase has not finished execution.
        """
        return self._avg_test_std_total_reward

    @property
    def min_test_std_total_reward(self) -> float or None:
        """
        The minimum standard deviation of total reward among all test volleys.
        It is None if test phase has not finished execution.
        """
        return self._min_test_std_total_reward

    @property
    def avg_test_std_scaled_reward(self) -> float or None:
        """
        The average standard deviation of scaled reward among all test volleys.
        It is None if test phase has not finished execution.
        """
        return self._avg_test_std_scaled_reward

    @property
    def min_test_std_scaled_reward(self) -> float or None:
        """
        The minimum standard deviation of scaled reward among all test volleys.
        It is None if test phase has not finished execution.
        """
        return self._min_test_std_scaled_reward

    @property
    def avg_test_avg_episode_length(self) -> int or None:
        """
        The average episode length among all test volleys.
        It is None if test phase has not finished execution.
        """
        return self._avg_test_avg_episode_length

    @property
    def validated(self) -> bool or None:
        """
        Flag defining if the experiment training is validated or not.
        It is None if experiment is not setup.
        """
        return self._validated

    @property
    def successful(self) -> bool or None:
        """
        Flag defining if the experiment testing is successful or not.
        It is None if experiment is not setup.
        """
        return self._successful



# Import packages

import tensorflow
import numpy
import scipy.signal
import random
import math

# Import required src

from usienarl import SpaceType, Config, Model
from usienarl.utils import softmax


class Buffer:
    """
    A buffer for storing trajectories experienced by a VPG agent interacting with the environment,
    and using Generalized Advantage Estimation (GAE-Lambda) for calculating the advantages of state-action pairs.

    The buffer is dynamically resizable and each sample stored is a numpy array of parallel samples.

    The buffer contains list of observations, actions, rewards, episode done flags, values (computed by the value stream
    of the model itself during prediction).
    When the buffer is emptied, the buffer also computes and returns advantages (using GAE) and rewards-to-go.
    """

    def __init__(self,
                 parallel_amount: int,
                 discount_factor: float, lambda_parameter: float):
        # Define buffer components
        self._observations: [] = []
        self._actions: [] = []
        self._rewards: [] = []
        self._values: [] = []
        self._last_step_rewards: [] = []
        # Define parameters
        self._discount_factor: float = discount_factor
        self._lambda_parameter: float = lambda_parameter
        self._parallel_amount: int = parallel_amount
        # Define buffer pointer attributes
        self._end_trajectories_pointers: [] = []
        self._pointer: numpy.ndarray = numpy.zeros(self._parallel_amount, dtype=int)
        self._episode_done_previous_step: numpy.ndarray = numpy.zeros(self._parallel_amount, dtype=bool)

    def store(self,
              observation: numpy.ndarray,
              action: numpy.ndarray,
              reward: numpy.ndarray,
              episode_done: numpy.ndarray,
              value: numpy.ndarray):
        """
        Store the time-step in the buffer, taking care of the parallelization of the environment.

        :param observation: the current observation to store in the buffer wrapped in a numpy array
        :param action: the last action to store in the buffer wrapped in a numpy array
        :param reward: the reward obtained from the action at the current state to store in the buffer wrapped in a numpy array
        :param episode_done: the flag defining the end of the episode at the current state to store in the buffer wrapped in a numpy array
        :param value: the value of the state as estimated by the value stream of the model to store in the buffer wrapped in a numpy array
        """
        # Append all data and increase the pointer
        self._observations.append(observation)
        self._actions.append(action)
        self._rewards.append(reward)
        self._values.append(value)
        # Update pointer by adding one where the episode is not already finished
        self._pointer = numpy.where(numpy.logical_not(self._episode_done_previous_step), self._pointer + 1, self._pointer)
        # Update the stored episode done flags
        self._episode_done_previous_step = episode_done.copy()

    def get(self) -> []:
        """
        Get all of the data from the buffer, serializing all the parallel data.
        Also reset pointers in the buffer and the lists composing the buffer.

        :return a list containing the numpy arrays of: observations, actions, advantages, rewards-to-go
        """
        # Generate numpy arrays out of the buffer
        observations_array: numpy.ndarray = numpy.stack(numpy.array(self._observations))
        actions_array: numpy.ndarray = numpy.stack(numpy.array(self._actions))
        # Prepare serialized list of all data in the buffer
        serialized_observations: [] = []
        serialized_actions: [] = []
        advantages: [] = []
        rewards_to_go: [] = []
        trajectory_start_pointer: numpy.ndarray = numpy.zeros(self._parallel_amount, dtype=int)
        for trajectory_index in range(len(self._end_trajectories_pointers)):
            pointer: numpy.ndarray = self._end_trajectories_pointers[trajectory_index]
            last_step_reward: numpy.ndarray = self._last_step_rewards[trajectory_index]
            for i in range(self._parallel_amount):
                # Get the trajectory slice for the current parallel episode
                trajectory_slice = slice(trajectory_start_pointer[i], pointer[i])
                # Compute rewards and values by appending the last step reward
                rewards: numpy.ndarray = numpy.array(numpy.array(self._rewards)[trajectory_slice, i].tolist() + [last_step_reward[i]])
                values: numpy.ndarray = numpy.array(numpy.array(self._values)[trajectory_slice, i].tolist() + [last_step_reward[i]])
                # Compute GAE-Lambda advantage estimation (compute advantages using the values in the buffer taken from the model)
                deltas: numpy.ndarray = rewards[:-1] + self._discount_factor * values[1:] - values[:-1]
                advantages += self._discount_cumulative_sum(deltas, self._discount_factor * self._lambda_parameter).tolist()
                # Compute rewards-to-go
                rewards_to_go += (self._discount_cumulative_sum(rewards, self._discount_factor)[:-1]).tolist()
                # Get the observations, actions and log likelihoods of the trajectory
                observations_trajectory: numpy.ndarray = observations_array[trajectory_slice]
                actions_trajectory: numpy.ndarray = actions_array[trajectory_slice]
                # Serialize all observations, actions and log likelihoods for all parallel episodes
                serialized_observations += observations_trajectory[:, i].tolist()
                serialized_actions += actions_trajectory[:, i].tolist()
            trajectory_start_pointer = pointer.copy()
        # Get the numpy arrays out of all serialized data (plus advantages and rewards to go, already serialized)
        observations_array = numpy.array(serialized_observations)
        actions_array = numpy.array(serialized_actions)
        rewards_to_go_array: numpy.ndarray = numpy.array(rewards_to_go)
        advantages_array: numpy.ndarray = numpy.array(advantages)
        # Execute the advantage normalization trick
        # Note: make sure mean and std are not zero!
        advantage_mean: float = float(numpy.mean(advantages_array)) + 1e-8
        global_sum_squared: float = float(numpy.sum((advantages_array - advantage_mean) ** 2)) + 1e-8
        advantage_std: float = numpy.sqrt(global_sum_squared / advantages_array.size) + 1e-8
        # Adjust advantages according to the trick
        advantages_array = ((advantages_array - advantage_mean) / advantage_std)
        # Reset the buffer
        self._observations = []
        self._actions = []
        self._rewards = []
        self._values = []
        self._last_step_rewards = []
        # Reset pointer
        self._pointer: numpy.ndarray = numpy.zeros(self._parallel_amount, dtype=int)
        self._end_trajectories_pointers: [] = []
        # Return all the buffer components
        return [observations_array, actions_array, advantages_array, rewards_to_go_array]

    def finish_trajectory(self,
                          last_step_reward: numpy.ndarray):
        """
        Finish the trajectory storing all pointers data related to each parallel episode and resetting flags.

        :param last_step_reward: the last reward given by the environment or the last predicted value if last state is not terminal
        """
        # Add the last step rewards
        self._last_step_rewards.append(last_step_reward)
        # Save the current pointer as a trajectory end pointer
        self._end_trajectories_pointers.append(self._pointer.copy())
        # Since the episode is done reset flags for previous step
        self._episode_done_previous_step: numpy.ndarray = numpy.zeros(self._parallel_amount, dtype=bool)

    @property
    def size(self) -> int:
        """
        The size of the buffer at the current time (it is dynamic).

        :return: the integer size of the buffer
        """
        return int(numpy.sum(self._pointer))

    @staticmethod
    def _discount_cumulative_sum(vector: numpy.ndarray, discount: float) -> numpy.ndarray:
        """
        Compute discounted cumulative sums of vectors.
        Credits to rllab.

        :param vector: the vector on which to compute cumulative discounted sum (e.g. [x0, x1, x2])
        :return the discounted cumulative sum (e.g. [x0 + discount * x1 + discount^2 * x2, x1 + discount * x2, x3])
        """
        return scipy.signal.lfilter([1], [1, float(-discount)], vector[::-1], axis=0)[::-1]


class VanillaPolicyGradient(Model):
    """
    Vanilla Policy Gradient with GAE (Generalized Advantage Estimation) buffer.
    The algorithm is on-policy and executes updates every a certain number of episodes. It is an actor-critic model.
    The model is constituted by two streams, the policy stream (the actor part of the model) adn the value stream (the
    critic part of the model). Hidden layers can be defined separately for each stream.
    The first stream computes and optimizes the policy loss. To drive the policy loss the advantages for each
    observation in the batch is required to be estimated. They are estimated by the buffer using GAE.
    The second stream computes the values of the current observations of the states and optimizes such estimation.
    To drive this loss, the value estimated by the stream itself for each state in the batch is required to be estimated.
    This estimation is done using the rewards. This stream is what makes the reward differentiable. Value update is run
    for a certain number of epochs.
    The buffer stores all the trajectories up to the update point. Since each episode can contains different numbers of
    steps, the buffer is dynamically resizable. When updated, the entire buffer is taken as batch and the buffer is
    cleared. This is necessary since to update the policy only trajectories collected using the current policy can be
    used. To stabilize learning and avoid overfitting, each update is done using minibatches. Usually actor and critic
    network are set to have the same structure.

    This algorithm is very likely to converge to local minimum and tends to be very unstable.

    Supported observation spaces:
        - discrete
        - continuous

    Supported action spaces:
        - discrete
        - continuous
    """

    def __init__(self,
                 name: str,
                 actor_hidden_layers_config: Config, critic_hidden_layers_config: Config,
                 discount_factor: float = 0.99,
                 learning_rate_policy: float = 3e-4, learning_rate_value: float = 1e-3,
                 value_update_epochs: int = 80,
                 minibatch_size: int = 32,
                 lambda_parameter: float = 0.97):
        # Define model attributes
        self.learning_rate_policy: float = learning_rate_policy
        self.learning_rate_value: float = learning_rate_value
        self.discount_factor: float = discount_factor
        # Define internal model attributes
        self._actor_hidden_layers_config: Config = actor_hidden_layers_config
        self._critic_hidden_layers_config: Config = critic_hidden_layers_config
        self._value_update_epochs: int = value_update_epochs
        self._minibatch_size: int = minibatch_size
        self._lambda_parameter: float = lambda_parameter
        # Define vanilla policy gradient empty attributes
        self.buffer: Buffer or None = None
        # Define internal vanilla policy gradient empty attributes
        self._observations = None
        self._actions = None
        self._actions = None
        self._advantages = None
        self._rewards = None
        self._mask = None
        self._logits = None
        self._masked_logits = None
        self._lower_bound = None
        self._upper_bound = None
        self._expected_value = None
        self._std = None
        self._log_std = None
        self._log_likelihood_actions = None
        self._log_likelihood_predictions = None
        self._value_predicted = None
        self._value_stream_loss = None
        self._policy_stream_loss = None
        self._value_stream_optimizer = None
        self._policy_stream_optimizer = None
        self._approximated_entropy = None
        # Generate the base model
        super(VanillaPolicyGradient, self).__init__(name)
        # Define the types of allowed observation and action spaces
        self._supported_observation_space_types.append(SpaceType.discrete)
        self._supported_observation_space_types.append(SpaceType.continuous)
        self._supported_action_space_types.append(SpaceType.discrete)
        self._supported_action_space_types.append(SpaceType.continuous)

    def _define_graph(self):
        # Set the GAE buffer for the vanilla policy gradient algorithm
        self.buffer: Buffer = Buffer(self._parallel, self.discount_factor, self._lambda_parameter)
        # Define the tensorflow model
        full_scope: str = self._scope + "/" + self._name
        with tensorflow.variable_scope(full_scope):
            # Define observations placeholder as an adaptable vector with shape Nx(O) where N is the number of examples and (O) the shape of the observation space
            # Note: it is the input of the model
            self._observations = tensorflow.placeholder(shape=(None, *self._observation_space_shape), dtype=tensorflow.float32, name="observations")
            # Define the actions placeholder as an adaptable vector with shape Nx(A) where N is the number of examples and (A) the shape of the action space
            self._actions = tensorflow.placeholder(shape=(None, *self._agent_action_space_shape), dtype=tensorflow.float32, name="targets")
            # Define the rewards placeholder as an adaptable vector of floats (they are actually rewards-to-go computed in the buffer)
            # Note: the model gets the rewards from the buffer
            self._rewards = tensorflow.placeholder(shape=(None,), dtype=tensorflow.float32, name="rewards")
            # Define the advantages placeholder as an adaptable vector of floats (computed with GAE in the buffer)
            # Note: the model gets the advantages from the buffer once computed using GAE on the values
            self._advantages = tensorflow.placeholder(shape=(None,), dtype=tensorflow.float32, name="advantages")
            # Define the policy stream
            # Note: this define the actor part of the model and its proper output (the predicted actions)
            with tensorflow.variable_scope("policy_stream"):
                # Define the policy stream network hidden layers from the config
                policy_stream_hidden_layers_output = self._actor_hidden_layers_config.apply_hidden_layers(self._observations)
                # Change the model definition according to its action space type
                if self._agent_action_space_type == SpaceType.discrete:
                    # Define the mask placeholder
                    self._mask = tensorflow.placeholder(shape=(None, *self._agent_action_space_shape), dtype=tensorflow.float32, name="mask")
                    # Define the logits as outputs with shape NxA where N is the size of the batch, A is the action size when its type is discrete
                    self._logits = tensorflow.layers.dense(policy_stream_hidden_layers_output, *self._agent_action_space_shape, activation=None, kernel_initializer=tensorflow.contrib.layers.xavier_initializer(), name="logits")
                    # Compute the masked logits using the given additive mask
                    self._masked_logits = tensorflow.add(self._logits, self._mask)
                    # Define the predicted actions on the first shape dimension as a squeeze on the samples drawn from a categorical distribution over the logits
                    self._actions_predicted = tensorflow.squeeze(tensorflow.multinomial(logits=self._masked_logits, num_samples=1), axis=1)
                    # Define the log likelihood according to the categorical distribution on actions given and actions predicted
                    self._log_likelihood_actions, _ = self.get_categorical_log_likelihood(self._actions, self._logits, name="log_likelihood_actions")
                    self._log_likelihood_predictions, _ = self.get_categorical_log_likelihood(tensorflow.one_hot(self._actions_predicted, depth=self._agent_action_space_shape[0]), self._logits, name="log_likelihood_predictions")
                else:
                    # Define the boundaries placeholders to clip the actions
                    self._lower_bound = tensorflow.placeholder(shape=(None, *self._agent_action_space_shape), dtype=tensorflow.float32, name="lower_bound")
                    self._upper_bound = tensorflow.placeholder(shape=(None, *self._agent_action_space_shape), dtype=tensorflow.float32, name="upper_bound")
                    # Define the expected value as the output of the deep neural network with shape Nx(A) where N is the number of inputs, (A) is the action shape when its type is continuous
                    self._expected_value = tensorflow.layers.dense(policy_stream_hidden_layers_output, *self._agent_action_space_shape, activation=None, kernel_initializer=tensorflow.contrib.layers.xavier_initializer(), name="expected_value")
                    # Define the log standard deviation and the standard deviation itself
                    self._log_std = tensorflow.get_variable(name="log_std", initializer=-0.5*numpy.ones(*self._agent_action_space_shape, dtype=numpy.float32))
                    self._std = tensorflow.exp(self._log_std, name="std")
                    # Define actions as the expected value summed up with a random gaussian vector multiplied by the standard deviation, clipped appropriately
                    self._actions_predicted = tensorflow.clip_by_value(self._expected_value + tensorflow.random_normal(tensorflow.shape(self._expected_value)) * self._std, self._lower_bound, self._upper_bound)
                    # Define the log likelihood according to the gaussian distribution on actions given and actions predicted
                    self._log_likelihood_actions = self.get_gaussian_log_likelihood(self._actions, self._expected_value, self._log_std, name="log_likelihood_actions")
                    self._log_likelihood_predictions = self.get_gaussian_log_likelihood(self._actions_predicted, self._expected_value, self._log_std, name="log_likelihood_predictions")
                # Define the policy stream loss as the mean of the advantages multiplied the log likelihood
                self._policy_stream_loss = -tensorflow.reduce_mean(self._advantages * self._log_likelihood_actions, name="policy_stream_loss")
                # Define the optimizer for the policy stream
                self._policy_stream_optimizer = tensorflow.train.AdamOptimizer(self.learning_rate_policy).minimize(self._policy_stream_loss)
            # Define the value stream
            # Note: this is the critic part of the model and the system making the reward differentiable (the value computation)
            with tensorflow.variable_scope("value_stream"):
                # Define the value stream network hidden layers from the config and its output (a single float value)
                value_stream_hidden_layers_output = self._critic_hidden_layers_config.apply_hidden_layers(self._observations)
                value_stream_output = tensorflow.layers.dense(value_stream_hidden_layers_output, 1, activation=None, kernel_initializer=tensorflow.contrib.layers.xavier_initializer(), name="value")
                # Define value by squeezing the output of the value stream
                self._value_predicted = tensorflow.squeeze(value_stream_output, axis=1, name="value_predicted")
                # Define value stream loss as the mean squared error of the difference between rewards-to-go given and the predicted value
                self._value_stream_loss = tensorflow.reduce_mean((self._rewards - self._value_predicted) ** 2, name="value_stream_loss")
                # Define the optimizer for the value stream
                self._value_stream_optimizer = tensorflow.train.AdamOptimizer(self.learning_rate_value).minimize(self._value_stream_loss)
            # Define approximated entropy for the logger
            self._approximated_entropy = tensorflow.reduce_mean(-self._log_likelihood_actions, name="approximated_entropy")
            # Define the initializer
            self._initializer = tensorflow.variables_initializer(tensorflow.get_collection(tensorflow.GraphKeys.GLOBAL_VARIABLES, full_scope), name="initializer")

    def sample_action(self,
                      session,
                      observation_current: numpy.ndarray,
                      possible_actions: [] = None):
        """
        Get the action sampled from the probability distribution of the model given the current observation and an optional mask.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon
        :param possible_actions: the optional list used to remove certain actions from the prediction (with discrete action space) or to bound the actions predicted (with continuous action space)
        :return: the action predicted by the model, with the value estimated at the current state and the relative log-likelihood of the sampled action
        """
        # Generate a one-hot encoded version of the observation if observation space type is discrete
        if self._observation_space_type == SpaceType.discrete:
            observation_current: numpy.ndarray = numpy.eye(*self._observation_space_shape)[observation_current.reshape(-1)]
        # Compute with respect of the action space shape
        if self._agent_action_space_type == SpaceType.discrete:
            # If there is no possible actions list and the action space type is discrete generate a full pass-through mask otherwise generate a mask upon it
            if possible_actions is None:
                mask: numpy.ndarray = numpy.zeros((self._parallel, *self._agent_action_space_shape), dtype=float)
            else:
                mask: numpy.ndarray = -math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
                for i in range(self._parallel):
                    mask[i, possible_actions[i]] = 0.0
            # Get action, value and log-likelihood with discrete action space shape
            action, value, log_likelihood = session.run([self._actions_predicted, self._value_predicted, self._log_likelihood_predictions],
                                                        feed_dict={
                                                            self._observations: observation_current,
                                                            self._mask: mask
                                                        })
        else:
            # If there is no possible action list and the action space type is continuous generate an unbounded range, otherwise use the given range
            if possible_actions is None:
                lower_bound: numpy.ndarray = -math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
                upper_bound: numpy.ndarray = math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
            else:
                lower_bound: numpy.ndarray = numpy.array(possible_actions)[:, 0]
                upper_bound: numpy.ndarray = numpy.array(possible_actions)[:, 1]
            # Get action, value and log-likelihood with continuous action space shape
            action, value, log_likelihood = session.run([self._actions_predicted, self._value_predicted, self._log_likelihood_predictions],
                                                        feed_dict={
                                                            self._observations: observation_current,
                                                            self._lower_bound: lower_bound,
                                                            self._upper_bound: upper_bound
                                                        })
        # Return the predicted action, the estimated value and the log-likelihood
        return action, value, log_likelihood

    def get_value_and_log_likelihood(self,
                                     session,
                                     action: numpy.ndarray,
                                     observation_current: numpy.ndarray,
                                     possible_actions: [] = None):
        """
        Get the estimated value of the given current observation and the log-likelihood of the given action.

        :param session: the session of tensorflow currently running
        :param action: the action of which to compute the log-likelihood
        :param observation_current: the current observation of the agent in the environment to estimate the value
        :param possible_actions: the optional list used to remove certain actions from the prediction (with discrete action space) or to bound the actions predicted (with continuous action space)
        :return: the value estimated at the current state and the log-likelihood of the given action
        """
        # Generate a one-hot encoded version of the observation if observation space type is discrete
        if self._observation_space_type == SpaceType.discrete:
            observation_current: numpy.ndarray = numpy.eye(*self._observation_space_shape)[observation_current.reshape(-1)]
        # Compute with respect of the action space shape
        if self._agent_action_space_type == SpaceType.discrete:
            # Generate a one-hot encoded version of the action if action space type is discrete
            action: numpy.ndarray = numpy.eye(*self._agent_action_space_shape)[action.reshape(-1)]
            # If there is no possible actions list and the action space type is discrete generate a full pass-through mask otherwise generate a mask upon it
            if possible_actions is None:
                mask: numpy.ndarray = numpy.zeros((self._parallel, *self._agent_action_space_shape), dtype=float)
            else:
                mask: numpy.ndarray = -math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
                for i in range(self._parallel):
                    mask[i, possible_actions[i]] = 0.0
            # Get value and log-likelihood with discrete action space shape
            value, log_likelihood = session.run([self._value_predicted, self._log_likelihood_actions],
                                                feed_dict={
                                                    self._observations: observation_current,
                                                    self._mask: mask,
                                                    self._actions: action
                                                })
        else:
            # If there is no possible action list and the action space type is continuous generate an unbounded range, otherwise use the given range
            if possible_actions is None:
                lower_bound: numpy.ndarray = -math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
                upper_bound: numpy.ndarray = math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
            else:
                lower_bound: numpy.ndarray = possible_actions[:, 0]
                upper_bound: numpy.ndarray = possible_actions[:, 1]
            # Get value and log-likelihood with continuous action space shape
            value, log_likelihood = session.run([self._value_predicted, self._log_likelihood_actions],
                                                feed_dict={
                                                    self._observations: observation_current,
                                                    self._lower_bound: lower_bound,
                                                    self._upper_bound: upper_bound,
                                                    self._actions: action
                                                })
        # Return the estimated value of the given current state and the log-likelihood of the given action
        return value, log_likelihood

    def get_value(self,
                  session,
                  observation_current: numpy.ndarray):
        """
        Get the estimated value of the given current observation.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to estimate the value
        :return: the value estimated at the current state
        """
        if self._observation_space_type == SpaceType.discrete:
            # Generate a one-hot encoded version of the observation if observation space type is discrete
            observation_current: numpy.ndarray = numpy.eye(*self._observation_space_shape)[observation_current.reshape(-1)]
        # Get the value predicted by the network at the current state
        value = session.run(self._value_predicted,
                            feed_dict={
                                self._observations: observation_current
                            })
        # Return the estimated value
        return value

    def get_log_likelihood(self,
                           session,
                           action,
                           observation_current: numpy.ndarray,
                           possible_actions: [] = None):
        """
        Get the the log-likelihood of the given action.

        :param session: the session of tensorflow currently running
        :param action: the action of which to compute the log-likelihood
        :param observation_current: the current observation of the agent in the environment
        :param possible_actions: the optional list used to remove certain actions from the prediction (with discrete action space) or to bound the actions predicted (with continuous action space)
        :return: the log-likelihood of the given action
        """
        # Generate a one-hot encoded version of the observation if observation space type is discrete
        if self._observation_space_type == SpaceType.discrete:
            observation_current: numpy.ndarray = numpy.eye(*self._observation_space_shape)[observation_current.reshape(-1)]
        # Compute with respect of the action space shape
        if self._agent_action_space_type == SpaceType.discrete:
            # Generate a one-hot encoded version of the action if action space type is discrete
            action: numpy.ndarray = numpy.eye(*self._agent_action_space_shape)[action.reshape(-1)]
            # If there is no possible actions list and the action space type is discrete generate a full pass-through mask otherwise generate a mask upon it
            if possible_actions is None:
                mask: numpy.ndarray = numpy.zeros((self._parallel, *self._agent_action_space_shape), dtype=float)
            else:
                mask: numpy.ndarray = -math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
                for i in range(self._parallel):
                    mask[i, possible_actions[i]] = 0.0
            # Get value and log-likelihood with discrete action space shape
            log_likelihood = session.run(self._log_likelihood_actions,
                                         feed_dict={
                                             self._observations: observation_current,
                                             self._mask: mask,
                                             self._actions: action
                                         })
        else:
            # If there is no possible action list and the action space type is continuous generate an unbounded range, otherwise use the given range
            if possible_actions is None:
                lower_bound: numpy.ndarray = -math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
                upper_bound: numpy.ndarray = math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
            else:
                lower_bound: numpy.ndarray = possible_actions[:, 0]
                upper_bound: numpy.ndarray = possible_actions[:, 1]
            # Get log-likelihood with continuous action space shape
            log_likelihood = session.run(self._log_likelihood_actions,
                                         feed_dict={
                                             self._observations: observation_current,
                                             self._lower_bound: lower_bound,
                                             self._upper_bound: upper_bound,
                                             self._actions: action
                                         })
        # Return the log-likelihood of the given action
        return log_likelihood

    def get_action_probabilities(self,
                                 session,
                                 observation_current: numpy.ndarray,
                                 possible_actions: [] = None) -> []:
        """
        Get all the action probabilities (softmax over masked logits if discrete, expected value and standard deviation if continuous) for the
        given current observation and an optional mask.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon
        :param possible_actions: the optional list used to remove certain actions from the prediction (with discrete action space) or to bound the actions predicted (with continuous action space)
        :return: the list of action probabilities (softmax over masked logits or expected values and std wrapped in a list depending on the agent action space type)
        """
        # Generate a one-hot encoded version of the observation if observation space type is discrete
        if self._observation_space_type == SpaceType.discrete:
            observation_current: numpy.ndarray = numpy.eye(*self._observation_space_shape)[observation_current.reshape(-1)]
        # Get the logits or the expected value and std of the distribution to compute the action probabilities depending on the action space shape
        if self._agent_action_space_type == SpaceType.discrete:
            # If there is no possible actions list and the action space type is discrete generate a full pass-through mask otherwise generate a mask upon it
            if possible_actions is None:
                mask: numpy.ndarray = numpy.zeros((self._parallel, *self._agent_action_space_shape), dtype=float)
            else:
                mask: numpy.ndarray = -math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
                for i in range(self._parallel):
                    mask[i, possible_actions[i]] = 0.0
            # Get logits on the current observation
            logits = session.run(self._masked_logits,
                                 feed_dict={
                                     self._observations: observation_current,
                                     self._mask: mask
                                 })
            # Return the softmax over the logits (probabilities of all actions)
            return softmax(logits)
        else:
            # If there is no possible action list and the action space type is continuous generate an unbounded range, otherwise use the given range
            if possible_actions is None:
                lower_bound: numpy.ndarray = -math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
                upper_bound: numpy.ndarray = math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
            else:
                lower_bound: numpy.ndarray = possible_actions[:, 0]
                upper_bound: numpy.ndarray = possible_actions[:, 1]
            # Get logits on the current observation
            expected_value, std = session.run([self._expected_value, self._std],
                                              feed_dict={
                                                  self._observations: observation_current,
                                                  self._lower_bound: lower_bound,
                                                  self._upper_bound: upper_bound
                                              })
            # Return the expected value and the standard deviation wrapped in a list
            return [expected_value, std]

    def update(self,
               session,
               batch: []):
        """
        Update the model weights (thus training the model) of the policy and value stream, given a batch of samples.
        Update is done through minibatches. Only one update of the policy (actor) is performed, while the value (critic)
        is updated multiple times.

        :param session: the session of tensorflow currently running
        :param batch: a batch of samples each one consisting in a tuple of observation, action, advantage and reward
        :return: the policy stream average loss of all minibatches, the value stream loss average over all updates and minibatches and the average approximated entropy of the policy
        """
        # Unpack the batch
        observations, actions, advantages, rewards = batch[0], batch[1], batch[2], batch[3]
        # Generate a one-hot encoded version of the observations if observation space type is discrete
        if self._observation_space_type == SpaceType.discrete:
            observations: numpy.ndarray = numpy.eye(*self._observation_space_shape)[numpy.array(observations).reshape(-1)]
        # Generate a one-hot encoded version of the actions if action space type is discrete
        if self._agent_action_space_type == SpaceType.discrete:
            actions: numpy.ndarray = numpy.eye(*self._agent_action_space_shape)[numpy.array(actions).reshape(-1)]
        # Run the policy optimizer of the model in training mode, also compute policy stream loss and approximated entropy
        policy_update_minibatch_iterations: int = 0
        policy_stream_loss_total: float = 0.0
        approximated_entropy_total: float = 0.0
        for minibatch in self._get_minibatch(observations, actions, advantages, rewards, self._minibatch_size):
            # Unpack the minibatch
            minibatch_observations, minibatch_actions, minibatch_advantages, _ = minibatch
            # Update the policy
            _, policy_stream_loss, approximated_entropy = session.run([self._policy_stream_optimizer, self._policy_stream_loss, self._approximated_entropy],
                                                                      feed_dict={
                                                                            self._observations: minibatch_observations,
                                                                            self._actions: minibatch_actions,
                                                                            self._advantages: minibatch_advantages,
                                                                      })
            policy_update_minibatch_iterations += 1
            policy_stream_loss_total += policy_stream_loss
            approximated_entropy_total += approximated_entropy
        policy_stream_loss_average: float = policy_stream_loss_total / policy_update_minibatch_iterations
        approximated_entropy_average: float = approximated_entropy_total / policy_update_minibatch_iterations
        # Run the value optimizer of the model in training mode for the required amount of steps, also compute average value stream loss
        value_stream_loss_average: float = 0.0
        for _ in range(self._value_update_epochs):
            value_update_minibatch_iterations: int = 0
            value_stream_loss_total: float = 0.0
            for minibatch in self._get_minibatch(observations, actions, advantages, rewards, self._minibatch_size):
                # Unpack the minibatch
                minibatch_observations, _, minibatch_advantages, minibatch_rewards = minibatch
                # Update the value
                _, value_stream_loss = session.run([self._value_stream_optimizer, self._value_stream_loss],
                                                   feed_dict={
                                                       self._observations: minibatch_observations,
                                                       self._advantages: minibatch_advantages,
                                                       self._rewards: minibatch_rewards
                                                   })
                value_update_minibatch_iterations += 1
                value_stream_loss_total += value_stream_loss
            # The average is only really saved on the last value update to know it at the end of all the update steps
            value_stream_loss_average = value_stream_loss_total / value_update_minibatch_iterations
        # Return all metrics
        return policy_stream_loss_average, value_stream_loss_average, approximated_entropy_average

    @property
    def warmup_steps(self) -> int:
        return 0

    @staticmethod
    def _get_minibatch(observations: [], actions: [], advantages: [], rewards: [],
                       minibatch_size: int) -> ():
        """
        Get a minibatch of the given minibatch size from the given batch (already unpacked).

        :param observations: the observations buffer in the batch
        :param actions: the actions buffer in the batch
        :param advantages: the advantages buffer in the batch
        :param rewards: the rewards buffer in the batch
        :param minibatch_size: the size of the minibatch
        :return: a tuple minibatch of shuffled samples of the given size
        """
        # Get a list of random ids of the batch
        batch_random_ids = random.sample(range(len(observations)), len(observations))
        # Generate the minibatches by shuffling the batch
        minibatch_observations, minibatch_actions, minibatch_advantages, minibatch_rewards = [], [], [], []
        for random_id in batch_random_ids:
            minibatch_observations.append(observations[random_id])
            minibatch_actions.append(actions[random_id])
            minibatch_advantages.append(advantages[random_id])
            minibatch_rewards.append(rewards[random_id])
            # Return the minibatch
            if len(minibatch_observations) % minibatch_size == 0:
                yield minibatch_observations, minibatch_actions, minibatch_advantages, minibatch_rewards
                # Clear the minibatch
                minibatch_observations, minibatch_actions, minibatch_advantages, minibatch_rewards = [], [], [], []

    @staticmethod
    def get_categorical_log_likelihood(actions_mask, logits, name: str):
        """
        Get log-likelihood for discrete action spaces (using a categorical distribution) on the given logits and with
        the actions mask.
        It uses tensorflow and as such should only be called in the define method.

        :param actions_mask: the actions used to mask the log-likelihood on the logits
        :param logits: the logits of the neural network
        :param name: the name of the tensorflow operation
        :return: the log-likelihood according to categorical distribution
        """
        # Define the unmasked likelihood as the log-softmax of the logits
        log_likelihood_unmasked = tensorflow.nn.log_softmax(logits)
        # Return the categorical log-likelihood by summing over the first axis of the target action mask multiplied
        # by the log-likelihood on the logits (unmasked, this is the masking operation) and the unmasked likelihood
        return tensorflow.reduce_sum(actions_mask * log_likelihood_unmasked, axis=1, name=name), log_likelihood_unmasked

    @staticmethod
    def get_gaussian_log_likelihood(actions, expected_value, log_std, name: str):
        """
        Get log-likelihood for continuous action spaces (using a gaussian distribution) on the given expected value and
        log-std and with the actions.
        It uses tensorflow and as such should only be called in the define method.

        :param actions: the actions used to compute the log-likelihood tensor
        :param expected_value: the expected value of the gaussian distribution
        :param log_std: the log-std of the gaussian distribution
        :param name: the name of the tensorflow operation
        :return: the log-likelihood according to gaussian distribution
        """
        # Define the log-likelihood tensor for the gaussian distribution on the given target actions
        log_likelihood_tensor = -0.5 * (((actions - expected_value) / (tensorflow.exp(log_std) + 1e-8)) ** 2 + 2 * log_std + numpy.log(2 * numpy.pi))
        # Return the gaussian log-likelihood by summing over all the elements in the log-likelihood tensor defined above
        return tensorflow.reduce_sum(log_likelihood_tensor, axis=1, name=name)



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import logging
import numpy
import tensorflow

# Import usienarl

from usienarl import Agent, Interface, SpaceType
from usienarl.models import DeepDeterministicPolicyGradient


class DDPGAgent(Agent):
    """
    Deep Deterministic Policy Gradient agent with simple Gaussian noise exploration policy.
    It is supplied with a DDPG model.

    The update every steps value define after how many steps an update should be performed. When updating, a number
    of updated is executed which is proportional to the number of elapsed steps.
    The noise is a gaussian white noise and it is necessary to perform exploration during training.
    """

    def __init__(self,
                 name: str,
                 model: DeepDeterministicPolicyGradient,
                 update_every_steps: int = 50,
                 summary_save_every_steps: int = 500,
                 batch_size: int = 100,
                 noise_scale_max: float = 0.1, noise_scale_min: float = 0.0,
                 noise_scale_decay: float = 1e-6):
        # Define internal attributes
        self._model: DeepDeterministicPolicyGradient = model
        self._update_every_steps: int = update_every_steps
        self._noise_scale_max: float = noise_scale_max
        self._noise_scale_min: float = noise_scale_min
        self._noise_scale_decay: float = noise_scale_decay
        self._summary_save_every_steps: int = summary_save_every_steps
        self._batch_size: int = batch_size
        # Define empty attributes
        self._last_update_step: int = 0
        self._last_summary_save_step: int = 0
        self._noise_scale: float or None = None
        # Generate base agent
        super(DDPGAgent, self).__init__(name)

    def _generate(self,
                  logger: logging.Logger,
                  observation_space_type: SpaceType, observation_space_shape: (),
                  agent_action_space_type: SpaceType, agent_action_space_shape: ()) -> bool:
        # Generate the model and check if it's successful, stop if not successful
        return self._model.generate(logger, self._scope + "/" + self._name,
                                    self._parallel,
                                    observation_space_type, observation_space_shape,
                                    agent_action_space_type, agent_action_space_shape)

    def initialize(self,
                   logger: logging.Logger,
                   session):
        # Reset attributes
        self._last_update_step = 0
        self._last_summary_save_step = 0
        # Initialize the model
        self._model.initialize(logger, session)
        # Initialize the noise scale value for exploration
        self._noise_scale = self._noise_scale_max
        # Run the weight copy operation to uniform main and target networks
        self._model.weights_copy(session)

    def act_warmup(self,
                   logger: logging.Logger,
                   session,
                   interface: Interface,
                   agent_observation_current: numpy.ndarray,
                   warmup_step: int, warmup_episode: int):
        # Act randomly
        return interface.sample_agent_action(logger, session)

    def act_train(self,
                  logger: logging.Logger,
                  session,
                  interface: Interface,
                  agent_observation_current: numpy.ndarray,
                  train_step: int, train_episode: int):
        # Get the possible actions (action boundaries)
        possible_actions: [] = interface.possible_agent_actions(logger, session)
        lower_bound: numpy.ndarray = possible_actions[:, 0]
        upper_bound: numpy.ndarray = possible_actions[:, 1]
        # Get the best action predicted by the model
        action = self._model.get_best_action(session, agent_observation_current, possible_actions)
        # Add a certain Gaussian noise according to its current scale
        for i in range(self._parallel):
            noise = self._noise_scale * numpy.random.randn(*self._agent_action_space_shape)
            action[i] += noise
        # Return the action clipped with the current possible action boundaries
        return numpy.clip(action, lower_bound, upper_bound)

    def act_inference(self,
                      logger: logging.Logger,
                      session,
                      interface: Interface,
                      agent_observation_current: numpy.ndarray,
                      inference_step: int, inference_episode: int):
        # Return the best action predicted by the model
        return self._model.get_best_action(session, agent_observation_current, interface.possible_agent_actions(logger, session))

    def complete_step_warmup(self,
                             logger: logging.Logger,
                             session,
                             interface: Interface,
                             agent_observation_current: numpy.ndarray,
                             agent_action: numpy.ndarray,
                             reward: numpy.ndarray,
                             episode_done: numpy.ndarray,
                             agent_observation_next: numpy.ndarray,
                             warmup_step: int, warmup_episode: int):
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current.copy(), agent_action.copy(), reward.copy(), agent_observation_next.copy(), episode_done.copy())

    def complete_step_train(self,
                            logger: logging.Logger,
                            session,
                            interface: Interface,
                            agent_observation_current: numpy.ndarray,
                            agent_action: numpy.ndarray,
                            reward: numpy.ndarray,
                            episode_done: numpy.ndarray,
                            agent_observation_next: numpy.ndarray,
                            train_step: int, train_episode: int):
        # Save the current step in the buffer
        self._model.buffer.store(agent_observation_current.copy(), agent_action.copy(), reward.copy(), agent_observation_next.copy(), episode_done.copy())
        # Execute update only after a certain number of steps each time
        if train_step - self._last_update_step >= self._update_every_steps:
            q_stream_loss_total: float = 0.0
            policy_stream_loss_total: float = 0.0
            # Update for as many step as you have been waiting
            for _ in range(train_step - self._last_update_step):
                # Execute the update and store policy and q-stream loss
                batch: [] = self._model.buffer.get(self._batch_size)
                q_stream_loss, policy_stream_loss = self._model.update(session, batch)
                # Update total metrics
                q_stream_loss_total += q_stream_loss
                policy_stream_loss_total += policy_stream_loss
                # Update target network
                self._model.weights_update(session)
            # Generate the tensorflow summary on the average metrics if required
            if self._summary_writer is not None:
                if train_step - self._last_summary_save_step >= self._summary_save_every_steps:
                    # Generate the tensorflow summary on the loss and add it to the writer
                    summary = tensorflow.Summary()
                    policy_stream_loss_average: float = policy_stream_loss_total / (train_step - self._last_update_step)
                    q_stream_loss_average: float = q_stream_loss_total / (train_step - self._last_update_step)
                    summary.value.add(tag="policy_stream_loss", simple_value=policy_stream_loss_average)
                    summary.value.add(tag="q_stream_loss", simple_value=q_stream_loss_average)
                    # Update the summary at the absolute current step
                    self._summary_writer.add_summary(summary, train_step)
                    # Update last summary save step
                    self._last_summary_save_step = train_step
            # Update last update step
            self._last_update_step = train_step

    def complete_step_inference(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                agent_observation_current: numpy.ndarray,
                                agent_action: numpy.ndarray,
                                reward: numpy.ndarray,
                                episode_done: numpy.ndarray,
                                agent_observation_next: numpy.ndarray,
                                inference_step: int, inference_episode: int):
        pass

    def complete_episode_warmup(self,
                                logger: logging.Logger,
                                session,
                                interface: Interface,
                                last_step_reward: numpy.ndarray,
                                episode_total_reward: numpy.ndarray,
                                warmup_step: int, warmup_episode: int):
        # Update the buffer at the end of trajectory
        self._model.buffer.finish_trajectory()

    def complete_episode_train(self,
                               logger: logging.Logger,
                               session,
                               interface: Interface,
                               last_step_reward: numpy.ndarray,
                               episode_total_reward: numpy.ndarray,
                               train_step: int, train_episode: int):
        # Update the buffer at the end of trajectory
        self._model.buffer.finish_trajectory()
        # Decrease the noise scale by its decay value
        self._noise_scale = max(self._noise_scale_min, self._noise_scale - self._noise_scale_decay)

    def complete_episode_inference(self,
                                   logger: logging.Logger,
                                   session,
                                   interface: Interface,
                                   last_step_reward: numpy.ndarray,
                                   episode_total_reward: numpy.ndarray,
                                   inference_step: int, inference_episode: int):
        pass

    @property
    def saved_variables(self):
        return self._model.trainable_variables

    @property
    def warmup_steps(self) -> int:
        return self._model.warmup_steps



#
# Copyright (C) 2019 Luca Pasqualini
# University of Siena - Artificial Intelligence Laboratory - SAILab
#
#
# USienaRL is licensed under a BSD 3-Clause.
#
# You should have received a copy of the license along with this
# work. If not, see <https://opensource.org/licenses/BSD-3-Clause>.

# Import packages

import os
import numpy
import logging
import matplotlib.pyplot as plot
import matplotlib.ticker as mticker
import enum

# Import required src

from usienarl import Environment, Volley, Agent, Interface


class EpisodeVolleyType(enum.Enum):
    """
    Enum type of the episode volley: training, validation or test.
    """
    training = 0
    validation = 1
    test = 2


class EpisodeVolley(Volley):
    """
    Episode based volley. Used for training, validation and test volleys. It is executed for a certain amount of
    episodes. When run in training and validation modes, plots of each episodes (averaged to always be 100 episodes)
    are saved.
    """
    def __init__(self,
                 environment: Environment,
                 agent: Agent,
                 interface: Interface,
                 parallel: int,
                 episode_volley_type: EpisodeVolleyType,
                 plots_path: str, plots_dpi: int,
                 episodes_required: int, episode_length: int):
        # Generate base volley
        super(EpisodeVolley, self).__init__(environment, agent, interface, parallel)
        # Make sure additional parameters are correct
        assert(episode_volley_type is not None)
        assert(plots_path is not None and plots_path)
        assert(plots_dpi > 0)
        assert(episodes_required > 0 and episode_length > 0)
        assert(parallel > 0)
        assert(episodes_required % parallel == 0)
        # Define internal attributes
        self._episode_volley_type: EpisodeVolleyType = episode_volley_type
        self._plots_path: str = plots_path
        self._plots_dpi: int = plots_dpi
        self._episodes_required: int = episodes_required
        self._episode_length: int = episode_length
        # Define empty attributes
        self._last_episode_done: numpy.ndarray or None = None
        self._last_reward: numpy.ndarray or None = None
        self._avg_total_reward: float or None = None
        self._avg_scaled_reward: float or None = None
        self._std_total_reward: float or None = None
        self._std_scaled_reward: float or None = None
        self._avg_episode_length: int or None = None
        self._rewards: [] = []
        self._total_rewards: [] = []
        self._scaled_rewards: [] = []
        self._episode_lengths: [] = []

    def _initialize(self) -> bool:
        # Reset empty attributes
        self._last_episode_done = None
        self._last_reward = None
        self._avg_total_reward = None
        self._avg_scaled_reward = None
        self._std_total_reward = None
        self._std_scaled_reward = None
        self._avg_episode_length = None
        self._rewards = []
        self._total_rewards = []
        self._scaled_rewards = []
        self._episode_lengths = []
        # This initialization always succeed
        return True

    def run(self,
            logger: logging.Logger,
            session,
            render: bool = False):
        # Print info
        if self._episode_volley_type == EpisodeVolleyType.training:
            logger.info("Training for " + str(self._episodes_required) + " episodes...")
        elif self._episode_volley_type == EpisodeVolleyType.validation:
            logger.info("Validating for " + str(self._episodes_required) + " episodes...")
        else:
            logger.info("Testing for " + str(self._episodes_required) + " episodes...")
        # Get the amount of parallel batches required
        parallel_episode_batches: int = self._episodes_required // self._parallel
        # Execute the parallel episode batches
        for parallel_episode_batch in range(parallel_episode_batches):
            # Print current progress every once in a while (if length is not too short)
            if parallel_episode_batches >= 100 and (parallel_episode_batch + 1) % (parallel_episode_batches // 10) == 0 and parallel_episode_batch > 0:
                if self._episode_volley_type == EpisodeVolleyType.training:
                    logger.info("Trained for " + str((parallel_episode_batch + 1) * self._parallel) + "/" + str(self._episodes_required) + " episodes...")
                elif self._episode_volley_type == EpisodeVolleyType.validation:
                    logger.info("Validated for " + str((parallel_episode_batch + 1) * self._parallel) + "/" + str(self._episodes_required) + " episodes...")
                else:
                    logger.info("Tested for " + str((parallel_episode_batch + 1) * self._parallel) + "/" + str(self._episodes_required) + " episodes...")
            # Initialize last reward and last episode done flags
            self._last_reward = numpy.nan * numpy.ones(self._environment.parallel, dtype=float)
            self._last_episode_done = numpy.zeros(self._environment.parallel, dtype=bool)
            # Execute actions until the all parallel step batches are completed or the maximum episode length is exceeded
            episode_rewards: [] = []
            state_current: numpy.ndarray = self._environment.reset(logger, session)
            for parallel_step_batch in range(self._episode_length):
                # Get the action decided by the agent
                observation_current: numpy.ndarray = self._interface.environment_state_to_observation(logger, session, state_current)
                if self._episode_volley_type == EpisodeVolleyType.training:
                    agent_action: numpy.ndarray = self._agent.act_train(logger, session, self._interface, observation_current,
                                                                        self._start_steps + self._steps, self._start_episodes + self._episodes)
                else:
                    agent_action: numpy.ndarray = self._agent.act_inference(logger, session, self._interface, observation_current,
                                                                            self._start_steps + self._steps, self._start_episodes + self._episodes)
                # Get the next state with relative reward and episode done flag
                environment_action: numpy.ndarray = self._interface.agent_action_to_environment_action(logger, session, agent_action)
                state_next, reward, episode_done = self._environment.step(logger, session, environment_action)
                # Send back information to the agent
                observation_next: numpy.ndarray = self._interface.environment_state_to_observation(logger, session, state_next)
                # Complete the step
                if self._episode_volley_type == EpisodeVolleyType.training:
                    self._agent.complete_step_train(logger, session, self._interface,
                                                    observation_current, agent_action, reward, episode_done, observation_next,
                                                    self._start_steps + self._steps, self._start_episodes + self._episodes)
                else:
                    self._agent.complete_step_inference(logger, session, self._interface,
                                                        observation_current, agent_action, reward, episode_done, observation_next,
                                                        self._start_steps + self._steps, self._start_episodes + self._episodes)
                # Render if required
                if render:
                    self._environment.render(logger, session)
                # Save the reward at the last step
                self._last_reward = numpy.where(episode_done * (1 - self._last_episode_done), reward, self._last_reward)
                if parallel_step_batch + 1 == self._episode_length:
                    self._last_reward = numpy.where(numpy.isnan(self._last_reward), reward, self._last_reward)
                # Add the reward to the list of rewards for this episode
                # Note: make sure the reward saved is NaN for all already completed episodes in the parallel batch
                episode_rewards.append(numpy.where(self._last_episode_done, numpy.nan, reward))
                # Update the current state with the previously next state
                state_current = state_next.copy()
                # Increase the number of trained steps
                # Note: the counter should be increased according to the completed episodes of the current parallel batch
                self._steps += numpy.count_nonzero(episode_done == 0)
                # Save volley steps at termination time
                # Note: saving the step of each final step of each volley is required to compute the average episode length
                step_array: numpy.ndarray = numpy.ones(self._environment.parallel, dtype=int) * (parallel_step_batch + 1)
                self._episode_lengths += step_array[episode_done * numpy.logical_not(self._last_episode_done)].tolist()
                if parallel_step_batch + 1 == self._episode_length:
                    self._episode_lengths += step_array[numpy.logical_not(episode_done)].tolist()
                # Increase the episode counter
                self._episodes += numpy.count_nonzero((episode_done * (1 - self._last_episode_done)) == 1)
                # Save the current episode done flags
                # Note: episode done flag is a numpy array so it should be copied
                self._last_episode_done = episode_done.copy()
                # Check if the episode is already completed
                if all(episode_done):
                    break
            # Consider done also all parallel episodes truncated but not terminated
            self._episodes += numpy.count_nonzero(self._last_episode_done == 0)
            # Save rewards per episode and total/scaled rewards
            self._rewards.append(episode_rewards.copy())
            self._total_rewards += numpy.nansum(numpy.array(episode_rewards), axis=0).tolist()
            self._scaled_rewards += numpy.nanmean(numpy.array(episode_rewards), axis=0).tolist()
            # Complete the episode and send back information to the agent
            if self._episode_volley_type == EpisodeVolleyType.training:
                self._agent.complete_episode_train(logger, session, self._interface,
                                                   self._last_reward, self._total_rewards[-1],
                                                   self._start_steps + self._steps, self._start_episodes + self._episodes)
            else:
                self._agent.complete_episode_inference(logger, session, self._interface,
                                                       self._last_reward, self._total_rewards[-1],
                                                       self._start_steps + self._steps, self._start_episodes + self._episodes)
        # Compute statistics
        self._avg_total_reward = numpy.round(numpy.average(numpy.array(self._total_rewards)), 3)
        self._avg_scaled_reward = numpy.round(numpy.average(numpy.array(self._scaled_rewards)), 3)
        self._std_total_reward = numpy.round(numpy.std(numpy.array(self._total_rewards)), 3)
        self._std_scaled_reward = numpy.round(numpy.std(numpy.array(self._scaled_rewards)), 3)
        self._avg_episode_length = numpy.rint(numpy.average(numpy.array(self._episode_lengths)))
        # Print results
        if self._episode_volley_type == EpisodeVolleyType.training:
            logger.info("Training for " + str(self._episodes_required) + " episodes finished with following result:")
        elif self._episode_volley_type == EpisodeVolleyType.validation:
            logger.info("Validating for " + str(self._episodes_required) + " episodes finished with following result:")
        else:
            logger.info("Testing for " + str(self._episodes_required) + " episodes finished with following result:")
        logger.info("Average total reward: " + str(self._avg_total_reward))
        logger.info("Standard deviation of total reward: " + str(self._std_total_reward))
        logger.info("Average scaled reward: " + str(self._avg_scaled_reward))
        logger.info("Standard deviation of scaled reward: " + str(self._std_scaled_reward))
        logger.info("Average episode length: " + str(self._avg_episode_length) + " steps")
        # Save the episodes plots
        self._save_plots(logger)

    def _save_plots(self,
                    logger: logging.Logger):
        """
        Save plots for each episode of the volley. If more than 100 episodes have been executed, an average of some
        interval of them is used to normalize to 100. This does nothing if volley type is test.
        """
        # If this is a test volley return
        if self._episode_volley_type == EpisodeVolleyType.test:
            return
        # Make sure all lengths are the same
        assert (len(self._total_rewards) == len(self._scaled_rewards) == len(self._episode_lengths))
        # Print info
        if self._episode_volley_type == EpisodeVolleyType.training:
            logger.info("Save training volley " + str(self._number) + " episodes plots...")
        else:
            logger.info("Save validation volley " + str(self._number) + " episodes plots...")
        # Compute statistic on the volley episodes
        # Note: if more than 100 episodes normalize to 100 episodes using interval averages (in order to avoid cluttering the plot)
        amount: int = len(self._total_rewards)
        interval: int = max(1, amount // 100)
        avg_total_rewards: [] = [sum(self._total_rewards[i:i + interval]) / interval for i in range(0, amount, interval)]
        avg_scaled_rewards: [] = [sum(self._scaled_rewards[i:i + interval]) / interval for i in range(0, amount, interval)]
        avg_episode_lengths: [] = [sum(self._episode_lengths[i:i + interval]) / interval for i in range(0, amount, interval)]
        # Generate volley dir:
        plot_directory = os.path.dirname(self._plots_path + "/volley_" + str(self._number) + "/")
        if not os.path.isdir(plot_directory):
            try:
                os.makedirs(plot_directory)
            except FileExistsError:
                pass
        # Save plots according to the requested type
        if self._episode_volley_type == EpisodeVolleyType.training:
            plot.plot(list(range(len(avg_total_rewards))), avg_total_rewards, 'r-')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            if interval > 1:
                plot.xlabel("Training episode (averaged every " + str(interval) + " episodes)")
            else:
                plot.xlabel("Training episode")
            plot.ylabel("Total reward")
            plot.savefig(plot_directory + "/v" + str(self._number) + "_training_episodes_total_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(avg_scaled_rewards))), avg_scaled_rewards, 'r--')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            if interval > 1:
                plot.xlabel("Training episode (averaged every " + str(interval) + " episodes)")
            else:
                plot.xlabel("Training episode")
            plot.ylabel("Scaled reward")
            plot.savefig(plot_directory + "/v" + str(self._number) + "_training_episodes_scaled_rewards.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(avg_episode_lengths))), avg_episode_lengths, 'b-.')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            if interval > 1:
                plot.xlabel("Training episode (averaged every " + str(interval) + " episodes)")
            else:
                plot.xlabel("Training episode")
            plot.ylabel("Episode length (steps)")
            plot.savefig(plot_directory + "/v" + str(self._number) + "_training_episodes_lengths_v.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
        else:
            plot.plot(list(range(len(avg_total_rewards))), avg_total_rewards, 'r-')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            if interval > 1:
                plot.xlabel("Validation episode (averaged every " + str(interval) + " episodes)")
            else:
                plot.xlabel("Validation episode")
            plot.ylabel("Total reward")
            plot.savefig(plot_directory + "/v" + str(self._number) + "_validation_episodes_total_rewards_v.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(avg_scaled_rewards))), avg_scaled_rewards, 'r--')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            if interval > 1:
                plot.xlabel("Validation episode (averaged every " + str(interval) + " episodes)")
            else:
                plot.xlabel("Validation episode")
            plot.ylabel("Scaled reward")
            plot.savefig(plot_directory + "/v" + str(self._number) + "_validation_episodes_scaled_rewards_v.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
            plot.plot(list(range(len(avg_episode_lengths))), avg_episode_lengths, 'b-.')
            plot.gca().xaxis.set_major_locator(mticker.MaxNLocator(integer=True))
            if interval > 1:
                plot.xlabel("Validation episode (averaged every " + str(interval) + " episodes)")
            else:
                plot.xlabel("Validation episode")
            plot.ylabel("Episode length (steps)")
            plot.savefig(plot_directory + "/v" + str(self._number) + "_validation_episodes_lengths_v.png", dpi=self._plots_dpi, transparent=True)
            plot.clf()
        # Print info
        if self._episode_volley_type == EpisodeVolleyType.training:
            logger.info("Plots of training volley " + str(self._number) + " saved successfully")
        else:
            logger.info("Plots of validation volley " + str(self._number) + " saved successfully")

    @property
    def last_episode_done(self) -> numpy.ndarray or None:
        """
        The episode done flag in the last step of the last episode of the volley.
        It is wrapped in a numpy array.
        It is None if volley is not setup.
        """
        return self._last_episode_done

    @property
    def last_reward(self) -> numpy.ndarray or None:
        """
        The reward in the last step of the last episode of the volley.
        It is wrapped in a numpy array.
        It is None if volley is not setup.
        """
        return self._last_reward

    @property
    def avg_total_reward(self) -> float or None:
        """
        The average total reward of all episodes of the volley.
        It is wrapped in a numpy array.
        It is None if volley is not setup or if volley has not finished execution.
        """
        return self._avg_total_reward

    @property
    def avg_scaled_reward(self) -> float or None:
        """
        The average scaled reward of all episodes of the volley.
        It is wrapped in a numpy array.
        It is None if volley is not setup or if volley has not finished execution.
        """
        return self._avg_scaled_reward

    @property
    def std_total_reward(self) -> float or None:
        """
        The standard deviation of total reward of all episodes of the volley.
        It is wrapped in a numpy array.
        It is None if volley is not setup or if volley has not finished execution.
        """
        return self._std_total_reward

    @property
    def std_scaled_reward(self) -> float or None:
        """
        The standard deviation of scaled reward of all episodes of the volley.
        It is wrapped in a numpy array.
        It is None if volley is not setup or if volley has not finished execution.
        """
        return self._std_scaled_reward

    @property
    def avg_episode_length(self) -> float or None:
        """
        The average episode length of all episodes of the volley.
        It is wrapped in a numpy array.
        It is None if volley is not setup or if volley has not finished execution.
        """
        return self._avg_episode_length

    @property
    def rewards(self) -> []:
        """
        The list of rewards per step grouped by episode of all episodes already executed.
        """
        return self._rewards

    @property
    def total_rewards(self) -> []:
        """
        The list of total rewards of all episodes already executed.
        """
        return self._total_rewards

    @property
    def scaled_rewards(self) -> []:
        """
        The list of scaled rewards of all episodes already executed.
        """
        return self._scaled_rewards

    @property
    def episode_lengths(self) -> []:
        """
        The list of episode lengths of all episodes already executed.
        """
        return self._episode_lengths



# Import packages

import tensorflow
import numpy
import math

# Import required src

from usienarl import SpaceType, Model, Config
from usienarl.utils import SumTree


class Buffer:
    """
    Prioritized experience replay buffer. It uses a sum tree to store not only the samples but also the priority of
    the samples, where the priority of the samples is a representation of the probability of a sample.

    When storing a new sample, the default priority value of the new node associated with such sample is set to
    the maximum defined priority. This value is iteratively changed by the algorithm when update is called.
    It automatically serializes all the parallel data stored in it.

    When getting the samples, a priority segment inversely proportional to the amount of required samples is generated.
    Samples are then uniformly taken for that segment, and stored in a minibatch which is returned.
    Also, a weight to compensate for the over-presence of higher priority samples is returned by the get method,
    along with the minibatch as second returned value.
    """

    _MINIMUM_ALLOWED_PRIORITY: float = 1.0
    _IMPORTANCE_SAMPLING_VALUE_UPPER_BOUND: float = 1.0
    _ABSOLUTE_ERROR_UPPER_BOUND: float = 1.0

    def __init__(self,
                 capacity: int,
                 parallel_amount: int,
                 minimum_sample_probability: float, random_sample_trade_off: float,
                 importance_sampling_value: float, importance_sampling_value_increment: float):
        # Define internal prioritized experience replay buffer attributes
        self._capacity: int = capacity
        self._parallel_amount: int = parallel_amount
        self._minimum_sample_probability: float = minimum_sample_probability
        self._random_sample_trade_off: float = random_sample_trade_off
        self._importance_sampling_value_increment: float = importance_sampling_value_increment
        self._importance_sampling_starting_value: float = importance_sampling_value
        self._sum_tree = SumTree(self._capacity)
        self._importance_sampling_value = self._importance_sampling_starting_value
        # Define internal prioritized experience replay buffer empty attributes
        self._sum_tree_last_sampled_indexes = None
        self._episode_done_previous_step: numpy.ndarray = numpy.zeros(self._parallel_amount, dtype=bool)

    def store(self,
              observation_current: numpy.ndarray,
              action: numpy.ndarray,
              reward: numpy.ndarray,
              observation_next: numpy.ndarray,
              episode_done: numpy.ndarray):
        """
        Store the time-step in the buffer, serializing immediately all the parallel episodes.

        :param observation_current: the current observation to store in the buffer wrapped in a numpy array
        :param action: the last action to store in the buffer wrapped in a numpy array
        :param reward: the reward obtained from the action at the current state to store in the buffer wrapped in a numpy array
        :param observation_next: the next observation to store in the buffer wrapped in a numpy array
        :param episode_done: whether or not this time-step was the last of the episode wrapped in a numpy array
        """
        # Serialize all the experiences to store in the buffer
        for i in range(self._parallel_amount):
            if self._episode_done_previous_step[i]:
                continue
            # Find the current max priority on the tree leafs
            max_priority: float = numpy.max(self._sum_tree.leafs)
            # If the max priority is zero set it to the minimum defined
            if max_priority <= 0:
                max_priority = self._MINIMUM_ALLOWED_PRIORITY
            # Set the max priority as the default one for this new sample
            # Note: we set the max priority for each new sample and then improve on it iteratively during training
            self._sum_tree.add((observation_current[i], action[i], reward[i], observation_next[i], episode_done[i]), max_priority)
        # Update the stored episode done flags
        self._episode_done_previous_step = episode_done.copy()

    def get(self,
            amount: int = 0):
        """
        Get a batch of data from the buffer of the given size. If size is not given all the buffer is used.

        :param amount: the batch size of data to get
        :return a list containing the arrays of: current observations, actions, rewards, next observations and episode done flags
        """
        # Adjust amount with respect to the size of the sum-tree
        if amount <= 0 or amount > self._sum_tree.size:
            amount = self._sum_tree.size
        # Define arrays of each sample components
        observations_current: [] = []
        actions: [] = []
        rewards: [] = []
        observations_next: [] = []
        episode_done_flags: [] = []
        # Define the returned arrays of indexes and weights
        self._sum_tree_last_sampled_indexes = numpy.empty((amount,), dtype=numpy.int32)
        importance_sampling_weights = numpy.empty((amount, 1), dtype=numpy.float32)
        # Get the segment of total priority according to the given amount
        # Note: it divides the sum tree priority by the amount and get the priority assigned to each segment
        priority_segment: float = self._sum_tree.total_priority / amount
        # Increase the importance sampling value of the defined increment value until the upper bound is reached
        self._importance_sampling_value = numpy.min((self._IMPORTANCE_SAMPLING_VALUE_UPPER_BOUND, self._importance_sampling_value + self._importance_sampling_value_increment))
        # Compute max importance sampling weight
        # Note: the weight of a given transition is inversely proportional to the probability of the transition stored
        # in the related leaf. The transition probability is computed by normalizing the priority of a leaf with the
        # total priority of the sum tree
        min_probability = numpy.min(self._sum_tree.leafs / self._sum_tree.total_priority)
        max_weight = (min_probability * amount) ** (-self._importance_sampling_value)
        # Return the sample
        for sample in range(amount):
            # Sample a random uniform value between the first and the last priority values of each priority segment
            lower_bound: float = priority_segment * sample
            upper_bound: float = priority_segment * (sample + 1)
            priority_value: float = numpy.random.uniform(lower_bound, upper_bound)
            # Get leaf index and related priority and data as stored in the sum tree
            leaf_index: int = self._sum_tree.get(priority_value)
            leaf_priority: float = self._sum_tree.get_priority(leaf_index)
            leaf_data = self._sum_tree.get_data(leaf_index)
            # Get the probability of the current sample
            sample_probability: float = leaf_priority / self._sum_tree.total_priority
            # Compute the importance sampling weights of each delta
            # The operation is: wj = (1/N * 1/P(j))**b / max wi == (N * P(j))**-b / max wi
            exponent: float = -self._importance_sampling_value
            importance_sampling_weights[sample, 0] = ((sample_probability * amount) ** exponent) / max_weight
            # Add the leaf index to the last sampled indexes list
            self._sum_tree_last_sampled_indexes[sample] = leaf_index
            # Generate the minibatch for this example
            observations_current.append(leaf_data[0])
            actions.append(leaf_data[1])
            rewards.append(leaf_data[2])
            observations_next.append(leaf_data[3])
            episode_done_flags.append(leaf_data[4])
        # Return the sample (minibatch) with related weights
        return [numpy.array(observations_current), numpy.array(actions), numpy.array(rewards), numpy.array(observations_next), numpy.array(episode_done_flags), importance_sampling_weights]

    def update(self,
               absolute_errors: []):
        """
        Update the buffer using the given absolute errors.

        :param absolute_errors: the absolute errors on the values predictions
        """
        # If no last sampled indexes are found, stop here
        if self._sum_tree_last_sampled_indexes is None:
            return
        # Avoid absolute error (delta) equal to zero (which would result in zero priority), by adding an epsilon
        absolute_errors += self._minimum_sample_probability
        # Force an upper bound of 1 on each absolute error (delta + epsilon)
        absolute_errors = numpy.minimum(absolute_errors, self._ABSOLUTE_ERROR_UPPER_BOUND)
        # Compute the priority to store as (delta + epsilon)^alpha
        priority_values = absolute_errors ** self._random_sample_trade_off
        # Get only the max priority values along each row (second axis)
        # Note: this is necessary since the absolute error is always zero between the current outputs and target outputs
        # when the action index is not the same of the chosen action of the sample
        priority_values = numpy.amax(priority_values, axis=1)
        # Before zipping reshape the absolute error array to be compatible with the stored tree indexes
        priority_values.reshape(self._sum_tree_last_sampled_indexes.shape)
        # For each last sampled sum tree leaf index and each correlated priority value update the sum tree
        for sum_tree_index, priority_value in zip(self._sum_tree_last_sampled_indexes, priority_values):
            self._sum_tree.update(sum_tree_index, priority_value)
        # Reset the last sampled sum tree indexes for the next update
        self._sum_tree_last_sampled_indexes = None

    def finish_trajectory(self):
        """
        Finish the trajectory, resetting episode done flags.
        """
        #  Reset stored episode done flags
        self._episode_done_previous_step: numpy.ndarray = numpy.zeros(self._parallel_amount, dtype=bool)

    @property
    def capacity(self) -> int:
        """
        The capacity of the buffer..

        :return: the integer capacity of the buffer
        """
        return self._capacity

    @property
    def size(self) -> int:
        """
        The size of the buffer at the current time.

        :return: the integer size of the buffer
        """
        return self._sum_tree.size


class Estimator:
    """
    Estimator defining the real Deep Expected SARSA model. It is used to define two identical models:
    target network and main-network.

    It is generated given the size of the observation and action spaces and the hidden layer config defining the
    hidden layers of the network.
    """

    def __init__(self,
                 scope: str,
                 observation_space_shape, agent_action_space_shape,
                 hidden_layers_config: Config,
                 error_clipping: bool = False,
                 huber_delta: float = 1.0):
        self.scope: str = scope
        with tensorflow.variable_scope(self.scope):
            # Define observations placeholder as a float adaptable array with shape Nx(O) where N is the number of examples and (O) the shape of the observation space
            self.observations = tensorflow.placeholder(shape=[None, *observation_space_shape], dtype=tensorflow.float32, name="observations")
            # Define the q-values targets placeholder with adaptable size NxA where N is the number of examples and A the size of the action space (always discrete)
            self.q_values_targets = tensorflow.placeholder(shape=[None, *agent_action_space_shape], dtype=tensorflow.float32, name="q_values_targets")
            # Define the estimator network hidden layers from the config
            hidden_layers_output = hidden_layers_config.apply_hidden_layers(self.observations)
            # Define the mask
            self.mask = tensorflow.placeholder(shape=[None, *agent_action_space_shape], dtype=tensorflow.float32, name="mask")
            # Define the estimator network predicted q-values given the input observations and the mask
            self.q_values_predictions = tensorflow.add(tensorflow.layers.dense(hidden_layers_output, *agent_action_space_shape, name="q_values_predictions"), self.mask)
            # Define the weights of the targets during the update process (e.g. the importance sampling weights)
            self.loss_weights = tensorflow.placeholder(shape=[None, 1], dtype=tensorflow.float32, name="loss_weights")
            # Define the absolute error
            self.absolute_error = tensorflow.abs(self.q_values_targets - self.q_values_predictions, name="absolute_error")
            # Define the loss with error clipping (huber loss) if required, otherwise mean square error loss
            if error_clipping:
                self.loss = tensorflow.reduce_mean(self.loss_weights * tensorflow.where(self.absolute_error < huber_delta,
                                                   0.5 * tensorflow.square(self.absolute_error),
                                                   self.absolute_error - 0.5), name="loss")
            else:
                self.loss = tensorflow.reduce_mean(self.loss_weights * tensorflow.square(self.absolute_error), name="loss")
            # Define the estimator weight parameters
            self.weight_parameters = [variable for variable in tensorflow.trainable_variables() if variable.name.startswith(self.scope)]
            self.weight_parameters = sorted(self.weight_parameters, key=lambda parameter: parameter.name)


class DeepExpectedSARSA(Model):
    """
    Deep Expected SARSA model with Expected SARSA update rule.
    The model is a deep neural network whose hidden layers can be defined by a config parameter.
    It uses a target network and a main network to correctly evaluate the expected future reward in order to stabilize
    learning. To further stabilize learning error can be clipped through the Huber Loss (note that this may cause
    instability in some environments).

    In order to synchronize the target network and the main network, every some interval steps model weights have to be
    copied from the main network to the target network.

    The update rule is the following (Bellman equation):
    Q(s, a) = R + gamma * mean_a(Q(s'))
    It uses the mean of the q-values estimated at the next state to update the estimate of the q-values at the current
    state.

    Supported observation spaces:
        - discrete
        - continuous

    Supported action spaces:
        - discrete
    """

    def __init__(self,
                 name: str,
                 hidden_layers_config: Config,
                 buffer_capacity: int,
                 learning_rate: float = 1e-3, discount_factor: float = 0.99,
                 minimum_sample_probability: float = 1e-2, random_sample_trade_off: float = 0.6,
                 importance_sampling_value: float = 0.4, importance_sampling_value_increment: float = 1e-3,
                 error_clipping: bool = False, huber_delta: float = 1.0):
        # Define model attributes
        self.learning_rate: float = learning_rate
        self.discount_factor: float = discount_factor
        # Define internal model attributes
        self._buffer_capacity: int = buffer_capacity
        self._minimum_sample_probability: float = minimum_sample_probability
        self._random_sample_trade_off: float = random_sample_trade_off
        self._importance_sampling_value: float = importance_sampling_value
        self._importance_sampling_value_increment: float = importance_sampling_value_increment
        self._hidden_layers_config: Config = hidden_layers_config
        self._error_clipping: bool = error_clipping
        self._huber_delta = huber_delta
        # Define model empty attributes
        self.buffer: Buffer or None = None
        # Define internal model empty attributes
        self._target_network: Estimator or None = None
        self._main_network: Estimator or None = None
        self._target_network_observations = None
        self._target_network_q_values_predictions = None
        self._main_network_observations = None
        self._main_network_q_values_predictions = None
        self._main_network_mask = None
        self._target_network_mask = None
        self._q_values_targets = None
        self._loss_weights = None
        self._absolute_error = None
        self._loss = None
        self._optimizer = None
        self._weight_copier = None
        # Generate the base model
        super(DeepExpectedSARSA, self).__init__(name)
        # Define the types of allowed observation and action spaces
        self._supported_observation_space_types.append(SpaceType.discrete)
        self._supported_observation_space_types.append(SpaceType.continuous)
        self._supported_action_space_types.append(SpaceType.discrete)

    def _define_graph(self):
        # Set the buffer
        self.buffer = Buffer(self._buffer_capacity, self._parallel,
                             self._minimum_sample_probability, self._random_sample_trade_off,
                             self._importance_sampling_value, self._importance_sampling_value_increment)
        # Define two estimator, one for target network and one for main network, with identical structure
        full_scope: str = self._scope + "/" + self._name
        self._main_network = Estimator(full_scope + "/MainNetwork",
                                       self._observation_space_shape, self._agent_action_space_shape,
                                       self._hidden_layers_config,
                                       self._error_clipping, self._huber_delta)
        self._target_network = Estimator(full_scope + "/TargetNetwork",
                                         self._observation_space_shape, self._agent_action_space_shape,
                                         self._hidden_layers_config,
                                         self._error_clipping, self._huber_delta)
        # Assign main and target networks to the model attributes
        self._main_network_observations = self._main_network.observations
        self._main_network_q_values_predictions = self._main_network.q_values_predictions
        self._target_network_q_values_predictions = self._target_network.q_values_predictions
        self._target_network_observations = self._target_network.observations
        self._main_network_mask = self._main_network.mask
        self._target_network_mask = self._target_network.mask
        self._q_values_targets = self._main_network.q_values_targets
        self._absolute_error = self._main_network.absolute_error
        self._loss = self._main_network.loss
        self._loss_weights = self._main_network.loss_weights
        # Define global operations
        with tensorflow.variable_scope(full_scope):
            # Define the optimizer
            self._optimizer = tensorflow.train.AdamOptimizer(self.learning_rate).minimize(self._loss)
            # Define the initializer
            self._initializer = tensorflow.variables_initializer(tensorflow.get_collection(tensorflow.GraphKeys.GLOBAL_VARIABLES, full_scope), name="initializer")
            # Define the weight copier operation (to copy weights from main network to target network)
            self._weight_copier = []
            for main_network_parameter, target_network_parameter in zip(self._main_network.weight_parameters, self._target_network.weight_parameters):
                copy_operation = target_network_parameter.assign(main_network_parameter)
                self._weight_copier.append(copy_operation)

    def get_q_values(self,
                     session,
                     observation_current: numpy.ndarray,
                     possible_actions: [] = None):
        """
        Get all the q-values according to the model at the given current observation.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon, wrapped in a numpy array
        :param possible_actions: the optional list used to remove certain actions from the prediction
        :return: all q-values predicted by the model
        """
        # If there is no possible actions list generate a full pass-through mask otherwise generate a mask upon it
        if possible_actions is None:
            mask: numpy.ndarray = numpy.zeros((self._parallel, *self._agent_action_space_shape), dtype=float)
        else:
            mask: numpy.ndarray = -math.inf * numpy.ones((self._parallel, *self._agent_action_space_shape), dtype=float)
            for i in range(self._parallel):
                mask[i, possible_actions[i]] = 0.0
        # Generate a one-hot encoded version of the observation if observation space is discrete
        if self._observation_space_type == SpaceType.discrete:
            observation_current: numpy.ndarray = numpy.eye(*self._observation_space_shape)[numpy.array(observation_current).reshape(-1)]
        # Compute the q-values predicted by main and target networks
        main_network_q_values = session.run(self._main_network_q_values_predictions,
                                            feed_dict={
                                                self._main_network_observations: observation_current,
                                                self._main_network_mask: mask
                                            })
        target_network_q_values = session.run(self._target_network_q_values_predictions,
                                              feed_dict={
                                                  self._target_network_observations: observation_current,
                                                  self._target_network_mask: mask
                                              })
        # Return the average of the q-values
        return (main_network_q_values + target_network_q_values) / 2

    def get_action_with_highest_q_value(self,
                                        session,
                                        observation_current: numpy.ndarray,
                                        possible_actions: [] = None):
        """
        Get the action with highest q-value from the q-values predicted by the model at the given current observation.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon, wrapped in a numpy array
        :param possible_actions: the optional list used to remove certain actions from the prediction
        :return: the action chosen by the model
        """
        # Return the action maximizing the predicted q-values given the current observation
        return numpy.argmax(self.get_q_values(session, observation_current, possible_actions), axis=1)

    def get_action_with_highest_q_value_and_q_values(self,
                                                     session,
                                                     observation_current: numpy.ndarray,
                                                     possible_actions: [] = None):
        """
        Get the action with highest q-value from the q-values predicted by the model at the given current observation
        and all the q-values according to the model at the given current observation.

        :param session: the session of tensorflow currently running
        :param observation_current: the current observation of the agent in the environment to base prediction upon, wrapped in a numpy array
        :param possible_actions: the optional list used to remove certain actions from the prediction
        :return: the action chosen by the model and all q-values predicted by the model
        """
        # Get q-values
        q_values = self.get_q_values(session, observation_current, possible_actions)
        # Return the highest q-value action and all the q-values
        return numpy.argmax(q_values, axis=1), q_values

    def copy_weights(self,
                     session):
        """
        Copy the weights from the main network to the target network.

        :param session: the session of tensorflow currently running
        """
        # Run all the weight copy operations
        session.run(self._weight_copier)

    def update(self,
               session,
               batch: []):
        """
        Update the model weights (thus training the model) using a batch of samples. Update is performed using the
        Expected SARSA Bellman equation to compute the model target q-values.

        :param session: the session of tensorflow currently running
        :param batch: a batch of samples each one consisting in a tuple of observation current, action, reward, observation next, episode done flag and sample weight
        :return: the loss and its relative absolute error
        """
        # Generate a full pass-through mask for each example in the batch
        masks: numpy.ndarray = numpy.zeros((len(batch[0]), *self._agent_action_space_shape), dtype=float)
        # Unpack the batch into numpy arrays
        observations_current, actions, rewards, observations_next, episode_done_flags, weights = batch[0], batch[1], batch[2], batch[3], batch[4], batch[5]
        # Define the input observations to the model (to support both space types)
        observations_current_input = observations_current
        observations_next_input = observations_next
        # Generate a one-hot encoded version of the observations if space type is discrete
        if self._observation_space_type == SpaceType.discrete:
            observations_current_one_hot: numpy.ndarray = numpy.eye(*self._observation_space_shape)[observations_current.reshape(-1)]
            observations_next_one_hot: numpy.ndarray = numpy.eye(*self._observation_space_shape)[observations_next.reshape(-1)]
            # Set the model input to the one-hot encoded representation
            observations_current_input = observations_current_one_hot
            observations_next_input = observations_next_one_hot
        # Get the q-values from the model at both current observations and next observations
        # Next observation is estimated by the target network
        q_values_current: numpy.ndarray = session.run(self._main_network_q_values_predictions,
                                                      feed_dict={
                                                          self._main_network_observations: observations_current_input,
                                                          self._main_network_mask: masks
                                                      })
        q_values_next: numpy.ndarray = session.run(self._target_network_q_values_predictions,
                                                   feed_dict={
                                                       self._target_network_observations: observations_next_input,
                                                       self._target_network_mask: masks
                                                   })
        # Apply Bellman equation with the Expected SARSA update rule
        for sample_index in range(len(actions)):
            # Extract current sample values
            action = actions[sample_index]
            reward: float = rewards[sample_index]
            episode_done: bool = episode_done_flags[sample_index]
            # Note: only the immediate reward can be assigned at end of the episode, i.e. when next observation is None
            if episode_done:
                q_values_current[sample_index, action] = reward
            else:
                q_values_current[sample_index, action] = reward + self.discount_factor * numpy.average(q_values_next[sample_index])
        # Train the model and save the value of the loss and of the absolute error as well as the summary
        _, loss, absolute_error = session.run([self._optimizer, self._loss, self._absolute_error],
                                              feed_dict={
                                                    self._main_network_observations: observations_current_input,
                                                    self._q_values_targets: q_values_current,
                                                    self._loss_weights: weights,
                                                    self._main_network_mask: masks
                                               })
        # Return the loss and the absolute error
        return loss, absolute_error

    @property
    def warmup_steps(self) -> int:
        return self._buffer_capacity
